﻿# -*- coding: utf-8 -*-

import bpy
from mathutils import Matrix

if(0):
	import importlib
	from .. import LWO_ParseHeader
	from . import Meshes, Bones, Curves
	from . import Opts
	importlib.reload(LWO_ParseHeader)
	importlib.reload(Meshes)
	importlib.reload(Bones)
	importlib.reload(Curves)
	importlib.reload(Opts)

from ..LWO_ParseHeader import *
from .. import FnCalc
from .Meshes import MakeMesh
from .Bones import MakeBone
from .Curves import MakeCurve
from .Opts import OptsObject

#### =======================================================================
__all__ = ("Build",)

#### =======================================================================
def __MakeMaters(surfs, clips, opts):
	u'''マテリアル作成'''
	loadedImgDct = {}
	mtlDct = {}
	if(opts.useExistMtl):
		for (kName, surf) in surfs.items():
			mt = bpy.data.materials.get(surf.name)
			if(not mt):		# 既設でなければ作成
				mt = __Mater(surf, clips, opts.useImgs, loadedImgDct)
			mtlDct[kName] = mt
	else:
		for (kName, surf) in surfs.items():
			mtlDct[kName] = __Mater(surf, clips, opts.useImgs, loadedImgDct)

	return(mtlDct)

#### --------------------------------
def __Mater(surf, clips, useImgs, loadedImgDct):
	mt = bpy.data.materials.new(surf.name)
	mt.diffuse_color = surf.colr + (1.0,)
	mt.specular_intensity = surf.spec
	mt.roughness = 0.0
	mt.use_nodes = True

	if(useImgs):
		for mp in surf.imap:
			if(mp.imgIdx):	# idx==0は未使用のフラグ
				if(mp.imgIdx not in loadedImgDct):
					im = __LoadImage(clips[mp.imgIdx])
					loadedImgDct[mp.imgIdx] = im

				else:
					im = loadedImgDct[mp.imgIdx]

				if(im != None):
					nd = mt.node_tree.nodes.new("ShaderNodeTexImage")
					nd.image = im
	return(mt)

#### --------------------------------
def __LoadImage(clip):
	if(clip.type == ClipT.STIL):
		return(__Image(clip.stillName, "FILE"))
	elif(clip.type == ClipT.ISEQ):
		return(__Image(clip.seqName, "SEQUENCE"))
	elif(clip.type == ClipT.ANIM):
		return(__Image(clip.animName, "MOVIE"))
	else:#	if(clip.type == "XREF"):
		pass

	return(None)

#### --------------------------------
def __Image(name, srcType):
	# .loadではファイル無し/不正パスでraiseするので空img作成後にパスを入れる(確認されない).
	nm = name.rsplit("/",1)[-1].rsplit("\\",1)[-1].rsplit(":",1)[-1].rsplit(".",1)[0]
	im = bpy.data.images.new(name=nm, width=32, height=32)
	im.source = srcType
	im.alpha_mode = "STRAIGHT"	# PREMUL
	im.colorspace_settings.name = "Raw"

	fpath = name[:]
	if((fpath[1] == ":") and ((fpath[2] != "\\") or (fpath[2] != "/"))):
		fpath = "{}/{}".format(fpath[0:2], fpath[2:])	# c:hoge.extが有り得る
	im.filepath = fpath

	return(im)

#### =======================================================================
def __PreCalcVertsBasePos(lyr, scale, usePivot):
	if(scale != (1.0,1.0,1.0)):
		if(usePivot):
			lyr.piv  = tuple(FnCalc.VMul(lyr.piv, scale))
			lyr.pnts = tuple( tuple(FnCalc.VMulSub(vec, scale, lyr.piv)) for vec in lyr.pnts )
		else:
			lyr.pnts = tuple( tuple(FnCalc.VMul(vec, scale)) for vec in lyr.pnts )
	else:
		if(usePivot):
			lyr.pnts = tuple( tuple(FnCalc.VSub(vec, lyr.piv)) for vec in lyr.pnts )
	if(usePivot):
		return((lyr.piv[0], lyr.piv[2], lyr.piv[1]))
	return((0.0, 0.0, 0.0))

#### =======================================================================
def Build(context, tags, layers, surfs, clips, opts= OptsObject()):
	u'''ブレンダーデータ作成'''

	deteminedMtlDct = __MakeMaters(surfs, clips, opts)	# 先行して作成

	if(not (opts.buildMesh | opts.buildCrv | opts.buildBone)):
		return

	buildedVertGrpsLsts = []
	buildedBonesDcts = {}

	lyrCnt = 0
	for lyr in layers:
		if(not (lyr.isHide & opts.ignoreHide)):
			obLoc = __PreCalcVertsBasePos(lyr, opts.scale, opts.usePivot)
			meOB = None

			if(opts.buildMesh and lyr.polNum):
				(meOB, buildedVertGrps) = MakeMesh(context, lyr, tags, surfs, deteminedMtlDct, opts, obLoc)
				buildedVertGrpsLsts.append(buildedVertGrps)

			if(opts.buildCrv and lyr.crvNum):
				MakeCurve(context, lyr, opts, obLoc)

			if(opts.buildBone and lyr.bonNum):
				(abOB, buildedBonesDct) = MakeBone(context, lyr, opts, obLoc)
				buildedBonesDcts.update({lyrCnt: buildedBonesDct})
				if(opts.assignAM and meOB):
					bpy.context.view_layer.objects.active = meOB
					bpy.ops.object.modifier_add(type="ARMATURE")
					meOB.modifiers["Armature"].object = abOB
		lyrCnt += 1

	if(opts.buildBone and opts.renWght and opts.useWght):
		lyrCnt = 0
		for lyr in layers:
			if((not (lyr.isHide & opts.ignoreHide)) and lyr.bonNum):
				for (k, bone) in lyr.bones.items():
					if(bone.weight):
						for (vgNames, vgs) in buildedVertGrpsLsts:
							try:
								idx = vgNames.index(bone.weight)
								vgs[idx].name = buildedBonesDcts[lyrCnt][k].name
								vgNames[idx] = None
							except:
								pass
			lyrCnt += 1

