﻿# -*- coding: utf-8 -*-

import bpy

from .. import FnCalc

#### =======================================================================
__all__ = ("MakeCurve",)

#### =======================================================================
def MakeCurve(context, lyr, opts, obLoc):
	u'''カーブ作成'''
	if(opts.crvPerLyr):
		obs = __MakeCurve_PerLyr(context, lyr, opts.mxAxisConv, opts.degree)
	else:
		obs = __MakeCurve_PerCrv(context, lyr, opts.mxAxisConv, opts.degree)

	for ob in obs:
		ob.hide_viewport = lyr.isHide	# 不可視だとEDITに入れなくなるので最後に.
		if(opts.usePivot):
			ob.location = obLoc

	return(obs)

#### --------------------------------
def __MakeCurve_PerLyr(context, lyr, mxAxisConv, deg):
	u'''カーブ作成, → 1 obj: 1 curve: N spline'''
	FnObjectLink = context.view_layer.active_layer_collection.collection.objects.link
	co = bpy.data.curves.new("{}_crv".format(lyr.name), "CURVE")
	ob = bpy.data.objects.new("{}_crv".format(lyr.name), co)
	FnObjectLink(ob)

	for crv in lyr.crvs:
		__MakeCurve_Base(co, crv, deg, lyr.pnts)
	co.transform(mxAxisConv)

	return([ob])

#### --------------------------------
def __MakeCurve_PerCrv(context, lyr, mxAxisConv, deg):
	u'''カーブ作成, → N obj: 1 curve: 1 spline'''
	FnObjectLink = context.view_layer.active_layer_collection.collection.objects.link
	obs = []
	cnt = 1
	for crv in lyr.crvs:
		co = bpy.data.curves.new("{}_crv{}".format(lyr.name, cnt), "CURVE")
		ob = bpy.data.objects.new("{}_crv{}".format(lyr.name, cnt), co)
		FnObjectLink(ob)
		obs.append(ob)

		__MakeCurve_Base(co, crv, deg, lyr.pnts)
		co.transform(mxAxisConv)
		cnt += 1

	return(obs)

#### --------------------------------
def __MakeCurve_Base(co, crv, deg, pnts):
	u'''カーブ作成'''
	co.dimensions = "3D"

	# crv=[flag, idx,idx,,,]
	if(crv[0]):	# Closed
		pntIdxs  = crv[2:-2]	# モデラカーブの補助点と始終の重複を除外
		isClosed = True
	else:		# Open
		pntIdxs  = crv[1:]
		isClosed = False

	sp = co.splines.new("NURBS")
	sp.points.add(len(pntIdxs )-1)	# 必要数-1, new段階で[0]が発生している.
	sp.points.foreach_set("co", tuple( v for i in pntIdxs for v in (pnts[i] + (1.0,)) ))	# (xyzw)
	sp.order_u = deg + 1			# 2~6,def:4
	sp.use_endpoint_u = True
	sp.use_cyclic_u = isClosed		# 周期時はendpnt指定は無視される
