﻿# -*- coding: utf-8 -*-

# print()の動作をv3スタイルに.
#from __future__ import print_function

import re

import maya.OpenMaya as om
import maya.OpenMayaAnim as oa
import maya.cmds as mc

import FnOM1, Meshes, Curves, Joints, VColors, Edges, Weights, Surfs
from .. import FnCalc
if(0):
	from .. import LWO_ParseHeader, LWO_Object
	import Defines
	reload(LWO_ParseHeader)
	reload(LWO_Object)
	reload(Defines)
	reload(Meshes)
	reload(Curves)
	reload(Joints)
	reload(VColors)
	reload(Edges)
	reload(Weights)
	reload(Surfs)
from ..LWO_ParseHeader import *
from ..LWO_Object import *
from .Defines import *
from .Opts import OptsObject

#### =======================================================================
__all__ = ("BuildObjects",)

#### ========================================================================
__gMeshDispColrChans = ("None", "Ambient", "Ambient+Diffuse", "Diffuse", "Specular", "Emission")

__DEBUG = 0

#### ========================================================================
def __Set_SelectionSet(dagpMesh, trsName, data, prefixFMT, buildedPolIdxs=None, isPnt=True):
	u'''{name:[idx,,,]} 型データ(SelectionSet/PICK,Ptag:COLR)を汎用Setに格納.'''
	FnSC = om.MFnSingleIndexedComponent()
	FnSt = om.MFnSet()
	selLst = om.MSelectionList()	# container 汎用

	componentType = om.MFn.kMeshVertComponent if(isPnt) else om.MFn.kMeshPolygonComponent

	selSet_Prefix = prefixFMT.format(NAME= re.sub("_mesh$|_ME$|_Mesh$", "", trsName))

	for (name, arElmIdxs) in data.iteritems():
		if(buildedPolIdxs):
			arElmIdxs = tuple( buildedPolIdxs[idx] for idx in arElmIdxs if(buildedPolIdxs[idx]!=None) )
		if(not arElmIdxs): continue

		moCmp = FnSC.create(componentType)		# FnSingle~へset, moCmpがカレントMObj
		FnSC.addElements(FnOM1.MIntArrayFromList(arElmIdxs))

		selLst.clear()				# 再利用, 前処理
		selLst.add(dagpMesh, moCmp)	# コンポーネント選択リスト作成

		selSetName = selSet_Prefix + name
		if(mc.objExists(selSetName)):	# 既存sets？
			FnSt.setObject(FnOM1.GetDependNode(FnOM1.MakeSelectionList(selSetName),0))	# 既存Setを受け取る
			FnSt.addMembers(selLst)
		else:							# Set作成/適用 // result強制表示2012~15, 消せない.
			oSet = FnSt.create(selLst, FnSt.kNone) # FnSt.kVerticesOnly 頂点縛りだとoutlinerに出ない
			FnSt.setName(selSetName)
			FnSt.setAnnotation("LWO SelectionSet")

#### ========================================================================
def __MakeMesh_MorfsNames_AddDict(baseName, moMesh, morfDict):
	buf = baseName.split(".", 1)	# [grp].[hoge.hoe]
	if(len(buf) == 1):				# 最大1分割,[0~1],len=1~2
		(grpName, morfName) = (MORF_NO_GRP, buf[0])
	else:
		(grpName, morfName) = (buf[0], buf[1])

	if(grpName in morfDict):	# grpの追加/新規, {grpName: {morf:MObj},{},...}
		morfDict[grpName].update({ morfName: moMesh })
	else:
		morfDict[grpName] = dict({ morfName: moMesh })

	return(grpName, morfName)

#### ------------------------------------------------------------------------
def __MakeMesh_AddGroupDict(grpName, FnTr, moTrsDict):
	if(grpName in moTrsDict):	# grpの追加/新規, {grpName: [MObj,MObj,...]}
		moTrsDict[grpName].append(FnTr.object())
	else:
		moTrsDict[grpName] = [FnTr.object()]

#### ------------------------------------------------------------------------
def __BlendShape_Mesh_ModifyRelFPA(fpaVt, idValDct, vScl):
	for (i, v) in idValDct.items():
		r = FnCalc.VMul(v, vScl)
		fpaVt[i].x += r[0]
		fpaVt[i].y += r[1]
		fpaVt[i].z += r[2]

#### ----------------
def __BlendShape_Mesh_ModifyAbsPivFPA(fpaVt, idValDct, vScl, piv):
	offs = FnCalc.VMul(piv, vScl)
	for (i, v) in idValDct.items():
		(fpaVt[i].x, fpaVt[i].y, fpaVt[i].z) = FnCalc.VMulSub(v, vScl, offs)

#### ----------------
def __BlendShape_Mesh_ModifyAbsFPA(fpaVt, idValDct, vScl):
	for (i, v) in idValDct.items():
		(fpaVt[i].x, fpaVt[i].y, fpaVt[i].z) = FnCalc.VMul(v, vScl)

#### ------------------------------------------------------------------------
def __BlendShape_Mesh(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, FnTrMorfGrp, morphs, isRelMorph, lyr, opts):
	FnTr = om.MFnTransform()	# 汎用
	FnMe = om.MFnMesh()			# 汎用
	dagp = om.MDagPath()		# 汎用

	(moTrsDict, morfDict) = ({}, {})
	for (baseName, idValDct) in sorted(morphs.items()):	# {id:pos, id:pos,...}
		# メッシュ作成, 未確定Meshへの変形と削除は不正処理される.
		# 元型create→copy→get/setPoints→deleteVertex == NG
		# 元型create→get/setPoints→deleteVertex == NG
		# PointArrayの複製を加工してcreate→deleteVertexする形にする
		fpaVt = om.MFloatPointArray(fpaVerts)
		if(isRelMorph):		# 相対
			__BlendShape_Mesh_ModifyRelFPA(fpaVt, idValDct, opts.vScl)
		else:
			if(opts.usePivot):
				__BlendShape_Mesh_ModifyAbsPivFPA(fpaVt, idValDct, opts.vScl, lyr.piv)
			else:
				__BlendShape_Mesh_ModifyAbsFPA(fpaVt, idValDct, opts.vScl)
		(dagpMesh, moMesh) = Meshes.MakeMesh_Build(newPolNum, buildedPolIdxs, fpaVt, miaPols, miaOrd, lyr, opts)

		# grpの追加/新規, {grpName: {morf:MObj},{},...}
		(grpName, morfName) = __MakeMesh_MorfsNames_AddDict(baseName, moMesh, morfDict)

		FnMe.setObject(moMesh)
		FnMe.setName(morfName + SFixNM.MF_SHAPE)	# 改名, 戻り名不要

		FnTr.setObject(dagpMesh.transform())		# 親Trs
		FnTr.setName(morfName + SFixNM.MF_TRS)		# 親Trsを改名, 戻り名不要

		# grpの追加/新規, {grpName: [MObj,MObj,...]}
		__MakeMesh_AddGroupDict(grpName, FnTr, moTrsDict)

		# mesh/Skelegon/curve混在対策, 未確定Meshの未使用頂点は一つ消せば全て消える
		if(lyr.bonNum):
#-			FnMe.deleteVertex(lyr.bones[0][BoneT.PNTID][0])	# 1本目の第1頂点のidx.
			FnMe.deleteVertex(lyr.bones[0].pnt0)	# 1本目の第1頂点のidx.
		elif(lyr.crvNum):
			FnMe.deleteVertex(lyr.crvs[0][1])	# 1本目の第1頂点のidx.

		FnMe.setObject(dagpMesh)	# dagpだとFix, Fixすると未使用頂点は消せない

	moGrpList = []		# morphGroup毎とlayer毎でグループ化
	APD = moGrpList.append
	for (grpName, moTrs) in moTrsDict.items():	# {grpName: [MObj,MObj,...]}
		APD(__ObjectGrouping(GrpVisT.SHOW, moTrs, "{}_grp".format(grpName)))

	tmpName = "{}Morph_grp".format("Rel" if(isRelMorph) else "Abs")
	mo = __ObjectGrouping(GrpVisT.SHOW, moGrpList, tmpName)	# Rel/Absグループ名の保証
	FnTrMorfGrp.addChild(mo)
	FnTr.setObject(mo)
	FnTr.setName(tmpName)

	return(morfDict)

#### ========================================================================
def __Get_MObject_ByName(names):
	mos = []
	APD = mos.append
	iter = FnOM1.MItSelectionList(FnOM1.MakeSelectionList(names))
	while(not iter.isDone()):
		APD(iter.GetDependNode())
		iter.next()
	return(mos)

#### ------------------------------------------------------------------------
def __BlendShape_Set(dagpMesh, morfDict):	# パラレル
	u''' '''
	# 頂点除外の為のset取得はMFnGeometryFilterから出来るはず.
	# 実用を考えるとBSノード自体が並行/平行でないと, NodeA=原型/移動量0, NodeB:変形有り
	# の様な時に破綻する. ノード上のターゲット間をパラレルにしても恩恵無いかも.
	# om.MFnでorder指定でのノード設置は出来ない. cmdsでやるしかない.
	FnBl = oa.MFnBlendShapeDeformer()
	FnMe = om.MFnMesh(dagpMesh)
	moMesh = FnMe.object()

	tmp = om.MDagPath(dagpMesh)
	tmp.pop()
	trsName = tmp.partialPathName()

	for (grpName, dct) in morfDict.items():
		bldName = trsName + grpName + SFixNM.BLSHAPE
		if(mc.objExists(bldName)):		# blendShapeが既存		※※※※※※※
			FnBl.setObject(__Get_MObject_ByName([bldName])[0])
			idxOffset = mc.getAttr("{}.w".format(bldName), s=True)
		else:
			FnBl.create(moMesh)			# ノード作成
			FnBl.setName(bldName)		# BlShapeにグループ名を付ける.
			idxOffset = 0

		for (i, (morfName, mo)) in enumerate(sorted(dct.items())):
			idx = i + idxOffset
			FnBl.addTarget(moMesh, idx, mo, 1.0)			# blShapeNodeは維持される
			# .weight[]のエリアス名が表示される. 初期のエリアス名はshape名.
			FnBl.setAlias(morfName, "weight[{}]".format(idx), FnBl.findPlug("weight"))	# [N]が必要.

#### --------
def __XXXBlendShape_Set(dagpMesh, morfDict):	# シーケンスのテスト
	FnBl = oa.MFnBlendShapeDeformer()
	mpElms = om.MIntArray()
	FnMe.setObject(dagpMesh)
	oMe = FnMe.object()

	tmp = om.MDagPath(dagpMesh)
	tmp.pop()
	trsName = tmp.partialPathName()

	for (grpName, dct) in morfDict.items():
		oBld = FnBl.create(oMe)

		num = len(dic.items())
		dif = (1000/num) * 0.001	# 少数3桁以上の精度を持つとNG
		wgt = 1.0					# DecrementでないとNG. Incrementだと3桁が維持出来ない？
		for (i, (morfName, mo)) in enumerate(sorted(dct.items())):
			FnBl.addTarget(oMe, 0, mo, wgt)				# blShapeNode(FnBlのmo)は維持される
			wgt -= dif									# ※wgt1からのDecrement
			# .weight[]のエリアス名が表示される.初期のエリアス名はshape名.
			mp = FnBl.findPlug("weight")				# 配列のルートを取る
			mp.getExistingArrayAttributeIndices(mpElms)	# Int配列[idx,idx,...]
			for c in mpElms:							# 有効な接続が[idx,...]に返ると期待して
				mpTgt = mp.elementByLogicalIndex(c)		# c番目のMPlug
				tgtName = mpTgt.name()					# Node.Attrなので分離
				attr = tgtName.split(".",1)				# 初期値と比較
				if(attr[1] == morfName + SFixNM.MF_SHAPE):
					FnBl.setAlias(morfName, "weight[{}]".format(i), mp)	# [N]が必要.
					break
		FnBl.setName(trsName + grpName + SFixNM.BLSHAPE)			# BlShape名にグループ名を付ける.

#### ------------------------------------------------------------------------
def __BlendShape_Group(name):
	FnTr = om.MFnTransform()
	FnTr.create()
	FnTr.setName(name + SFixNM.MF_GRP)	# 戻り名不要
	FnTr.findPlug("v", False).setBool(False)
	return(FnTr)

#### ------------------------------------------------------------------------
def __MakeMesh_MorfTargetMesh(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, lyr, opts):
	(hasRM, hasAM) = (len(lyr.relMorphs), len(lyr.absMorphs))
	(relMorfDict, absMorfDict) = ({}, {})
	if(opts.loadRM or opts.loadAM):	# モーフジオメトリ作成ならオブジェクトと対になるグループを作る
		if(__DEBUG): print("CreateMorph:")

		if(hasRM + hasAM):
			FnTrMorfGrp = __BlendShape_Group(lyr.name)
		if(opts.loadRM and hasRM):
			relMorfDict = __BlendShape_Mesh(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, FnTrMorfGrp, lyr.relMorphs, MorfT.REL, lyr, opts)
		if(opts.loadAM and hasAM):
			absMorfDict = __BlendShape_Mesh(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, FnTrMorfGrp, lyr.absMorphs, MorfT.ABS, lyr, opts)
	return(hasRM, hasAM, relMorfDict, absMorfDict)

#### ------------------------------------------------------------------------
def __SetAttrToBaseMesh_withUpdateSurf(moMesh, hasVertColor, opts):
	FnMe = om.MFnMesh(moMesh)
	# カラーチャネル dcc/displayColorChannel
	# "None", "Ambient", "Ambient+Diffuse", "Diffuse", "Specular", "Emission"
	FnMe.findPlug("dcc", False).setString(__gMeshDispColrChans[3])
	# 両面 doubleSided
	FnMe.findPlug("ds", False).setBool(opts.isDblSide)
	# .dcol/displayColors
	if(hasVertColor != None): FnMe.setDisplayColors(hasVertColor)

	FnMe.updateSurface()

#### ========================================================================
def MakeMeshObject(tags, lyr, surfs, clips, opts):
	if((not lyr.pntNum) or (not lyr.polNum)): return((None,)*4)

	FMT = u"Geometry 構築 ({}: {}, Scale: {:g}, PivotOffs: {:g}, {:g}, {:g})"
	print(FMT.format(lyr.idx +1, lyr.name, opts.scl, *lyr.piv))

	if(__DEBUG): print("CreateMesh:")

	(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd) = Meshes.MakeMesh_Collection(lyr, opts)
	(dagpMesh, moMesh) = Meshes.MakeMesh_Build(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, lyr, opts)

	# UVセット作成とカレントをセット
	Meshes.MakeMesh_SetUVs(moMesh, newPolNum, buildedPolIdxs, lyr)

	# エッジウェイト適用
	if(opts.useEdge): Edges.ApplyCreaseEdge(moMesh, lyr)

	# perPolノーマル適用

	# 頂点カラー
	(rgbNames, rgbaNames, hasVertColor) = ([], [], None)
	if(opts.useVCol):
		# 頂点カラー適用
		rgbNames  = VColors.ApplyVColors(moMesh, buildedPolIdxs, lyr.pols, lyr.rgb,  opts.useMiVC)
		rgbaNames = VColors.ApplyVColors(moMesh, buildedPolIdxs, lyr.pols, lyr.rgba, opts.useMiVC)
		# 頂点カラー表示
		hasVertColor = True if(rgbNames or rgbaNames) else False

	# サーフェス(SG)適用
	(mtNames, sgNames) = Surfs.MakeSurface(tags, lyr.surfTags, buildedPolIdxs, dagpMesh, opts)
	# マテリアルの設定
	Surfs.SetSurfaceAttr(surfs, clips, mtNames, sgNames, rgbNames, rgbaNames, opts)

	# セレクションセット作成
	if(opts.usePick): __Set_SelectionSet(dagpMesh, lyr.name, lyr.picks, SetFMT.SELS, None, True)
	if(opts.useColr): __Set_SelectionSet(dagpMesh, lyr.name, lyr.tagColors, SetFMT.COLR, buildedPolIdxs, False)
	if(opts.usePart): __Set_SelectionSet(dagpMesh, lyr.name, lyr.tagParts, SetFMT.PART, buildedPolIdxs, False)

	# レイヤがPoly/Skelegon/Curve混在の場合、頂点が未使用頂点としてmesh内に残る.
	# 削除によるindexズレ(uv等)対策をするのではなく、ジオメトリ構築後に削除する.
	# [ 1点でも削除すれば全ての未使用頂点が消え、indexが詰められる.]
	# MayaUI操作ではエッジの削除等でトポロジ変更を起こせば消える(頂点単位の明示的削除は出来ない).
	if(lyr.bonNum):	# レイヤ内にボーンがある.
		FnMe = om.MFnMesh(moMesh)
#-		FnMe.deleteVertex(lyr.bones[0][BoneT.PNTID][0])	# 1本目の第1頂点のidx.
		FnMe.deleteVertex(lyr.bones[0].pnt0)	# 1本目の第1頂点のidx.

		if(len(lyr.weights) and (opts.useWgt or opts.useXML)):	# ボーンが参照するvmap値を保持/書き出しする場合.
			Weights.MakeWeightMap(dagpMesh.fullPathName(), lyr.bones, lyr.weights, lyr.pnts, lyr.pols, buildedPolIdxs, opts)

	elif(lyr.crvNum):
		FnMe = om.MFnMesh(moMesh)
		FnMe.deleteVertex(lyr.crvs[0][1])	# 1本目の第1頂点のidx.

	FnMe = om.MFnMesh(dagpMesh)	# dagpだとFix, Fixすると未使用頂点は消せない

	# モーフ形状作成
	(hasRM, hasAM, relMorfDict, absMorfDict) = __MakeMesh_MorfTargetMesh(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, lyr, opts)

	# ブレンドシェイプ作成
	if(opts.useRM and opts.loadRM and hasRM): __BlendShape_Set(dagpMesh, relMorfDict)
	if(opts.useAM and opts.loadAM and hasAM): __BlendShape_Set(dagpMesh, absMorfDict)

	__SetAttrToBaseMesh_withUpdateSurf(moMesh, hasVertColor, opts)

	return(("",)*4)

#### ========================================================================
def __ObjectGrouping(isVis, arTrs, name):
	u'''grp用transformを作り,arTrs内のmoを子にする'''
	if(not arTrs): return(None)		# moでなくNoneで.

	FnTrs = om.MFnTransform()
	FnTrs.create()
	FnTrs.setName(name)
	if(isVis == GrpVisT.HIDE):
		FnTrs.findPlug("v", False).setBool(isVis)

	for mo in arTrs:
		FnTrs.addChild(mo)
	return(FnTrs.object())	# MObjectを返す

#### ========================================================================
def BuildObjects(tags, layers, surfs, clips, **kwd):
	opts = kwd.get("o", kwd.get("opts", None))
	if((not opts) or (not isinstance(opts, OptsObject))):
		opts = OptsObject()

	# actListはfile内格納順に応じたboolのリスト, 便宜上,[空]も認める
	actList = kwd.get("al", kwd.get("actList", None))
	if(actList == None):	# LWO単位での選択, 指定とフラグからactListを作成
		actList = tuple( (not (lyr.isHide & opts.skipHide)) for lyr in layers )
#		actList = map(lambda lyr: (not (lyr.isHide & opts.skipHide)), layers)

	else:					# Layer単位での選択, 保険として比較
		if(len(layers) != len(actList)):
			mc.warning(u"GUIとファイルでレイヤ数が不一致"),
			return

	for (lyr, isActive) in zip(layers, actList):
		if(not isActive): continue

		if(opts.isMesh):
			MakeMeshObject(tags, lyr, surfs, clips, opts)

		if(opts.isCrv):
			arTrs = Curves.MakeCurveObject(lyr, opts.vScl, opts.deg)
			__ObjectGrouping(GrpVisT.SHOW, arTrs, "{}_Curve_grp".format(lyr.name))

		if(opts.isBone):
			arTrs = Joints.MakeJointObject(lyr, opts.vScl, opts.isZAxis)
			__ObjectGrouping(GrpVisT.SHOW, arTrs, "{}_Joint_grp".format(lyr.name))
