﻿# -*- coding: utf-8 -*-

import maya.OpenMaya as om

from .. import FnCalc
from .Defines import *
from .Opts import OptsObject	# .DEFALUT_XXを参照してる

#### ========================================================================
__all__ = ("MakeCurveObject",)

#### ========================================================================
"""
def __MakeCurve_CpStdKnots(cvNum, deg):
	u'''開いたカーブをcpで作る時のknot'''
	k = ([0.0]*(deg-1)) + [float(i) for i in range(cvNum-deg)] + ([float(cvNum-deg)]*deg))
	return(FnOM1.MDoubleArrayFromList(k))
"""

#### ========================================================================
def MakeCurveObject(lyr, vScl=[OptsObject.DEFAULT_SCL,OptsObject.DEFAULT_SCL,-OptsObject.DEFAULT_SCL], deg=OptsObject.DEFAULT_DEG):
	if(lyr.crvNum == 0): return([])

	baseName = lyr.name.replace(" ", "_")	# maya自体の変換もあるけど呪い
	if(lyr.crvNum == 1):
		names = (NameFMT.CRVNAME.format(NAME= baseName),)
	else:
		FMT = NameFMT.CRVNAMES.format(NAME= baseName, SPC= u"_" if(baseName[-1].isdigit()) else "")
		names = tuple( FMT.format(CNT= i) for i in range(1, lyr.crvNum+1) )

	FnCrv= om.MFnNurbsCurve()
	FnDag= om.MFnDagNode()
	mpa  = om.MPointArray()

	arTrs = []
	for (name, crv) in zip(names, lyr.crvs):
		# crv[flag, idx,idx,,,]
		if(crv[0]):	# [0]Flag
			form = FnCrv.kPeriodic
			isUniformParam = True
			crvPnts = crv[2:-1]
			# 前後の1点づつ落とし, 循環状態にしてからepで作る.
		else:		# 
			form = FnCrv.kOpen
			isUniformParam = False
			crvPnts = crv[1:]

		mpa.clear()	# 再利用, 前処理
		mpa.setLength(len(crvPnts))

		for (i, v) in enumerate(crvPnts):
			mpa.set(i, *FnCalc.VMul(lyr.pnts[v], vScl))

		moTrs = FnCrv.createWithEditPoints(mpa, deg, form, False, False, isUniformParam)

		FnDag.setObject(moTrs)
		FnDag.setName(name)
		FnCrv.setName("{}Shape".format(name))

		arTrs.append(moTrs)

	return(arTrs)
