﻿# -*- coding: utf-8 -*-

import os, re
from functools import partial

import maya.cmds as mc

from .. import ParseLWO
import BuildLWO, ApplyWeightsByAttr
if(0):
	reload(ParseLWO)
	reload(BuildLWO)
	reload(ApplyWeightsByAttr)

#### =======================================================================
__all__ = ("LoadLWO")

#### =======================================================================
# form文字定数 -----------
class FPos(object):
	TP = "top"
	BT = "bottom"
	LT = "left"
	RT = "right"
	__slots__ = ()

# popupMenu(button=) -----------
class PClick(object):
	RGT = 0
	LFT = 1
	MID = 2
	__slots__ = ()

# LabelFont -----
class LblFont(object):
	PLN   = "plainLabelFont"		# プロポーショナル
	PLN_S = "smallPlainLabelFont"
	BLD   = "boldLabelFont"			# defaultFont
	BLD_S = "smallBoldLabelFont"
	TIN   = "tinyBoldLabelFont"
	FIX   = "fixedWidthFont"		# 固定幅
	FIX_S = "smallFixedWidthFont"
	OBL   = "obliqueLabelFont"		# 斜体
	OBL_S = "smallObliqueLabelFont"
	__slots__ = ()

# IconStyle (iconTextCB/RadBtn(st=),)-----
class ICStyle(object):
	ICON    = "iconOnly"
	TEXT    = "textOnly"
	HORIZON = "iconAndTextHorizontal"
	VERTICAL= "iconAndTextVertical"
	CENTER  = "iconAndTextCentered"
	__slots__ = ()

# 空object -----
class BaseObject(object):
	def __setitem__(self, k,v):		# obj[k]=v
		self.__dict__[k] = v
	def __getitem__(self, k):		# =obj[k]
		return(self.__dict__[k])

# BGC object -----
class BGCols(object):
	(DEF, BASE, HCON , BG) = (0.27, 0.32, 0.42, 0.1647)
	DEF3 = (DEF,)*3
	BASE3= (BASE,)*3
	HCON3= (HCON,)*3
	BG3  = (BG,)*3
	MPOP_GREEN = (BASE, BASE+.15, BASE+.1)
	LPOP_BLUE  = (BASE, BASE+.1,  BASE+.15)
	__slots__ = ()

#### =======================================================================
class LoadLWO(object):
	__slots__ = (
		"ctls",			# UI ctl
		"subStore",		# 履歴リスト
		"presetPath",	# プリセットpathリスト
		"lyrVisList",	# 時点のレイヤ可視属性リスト
		"tmpSelList",	# レイヤ選択On/Off時のレイヤ選択状態退避用
		"dialogPath",	# 前回取得時のpath
		"wndName",
		"isDebug",
		"isVerb",
		"flags",
	)

	TEMPLA_BASE = "LOAD_LWO_PANEL_TEMPLATE_BASE"	# UI template名

	WINDOWNAME  = "Load_LWO_panels"

	FILE_CAPTION = "Load LWO"
	FILE_FILTER  = u"Lightwave Files (*.lwo);;全てのファイル (*.*)"
	FILE_EXT     = ".lwo"

	#### ----------------
	class CSize(object):
		RHGT = 20
		FHGT = 18
		SCL  = 50
		CRV  = 35
		SHD  = 60
		W    = 104
		Wh   = W / 2
		__slots__=()

	#### ----------------
	class ShdType(object):	# Mtl Choice Idx, UI用故1から.
		LAMB = 1
		BLIN = 2
		PHON = 3
		SURF = 4
		__slots__=()

	__gShadLbls = (u"Lambert", u"Blinn", u"Phong")
	__gShadAlign= (FPos.LT,)*3
	__gShadWidth= (65, CSize.SHD, CSize.SHD)
	__gColW_Shader = (5, CSize.W, CSize.W*3)

	__gColW_Base  = (5, CSize.Wh, CSize.Wh, CSize.W*2, CSize.W)
	__gColW_PolMakes = (5,) + (CSize.W,)*4
	__gColW_Sets  = (4,) + (CSize.W,)*4
	__gColW_Image = (5, CSize.W, CSize.Wh, CSize.Wh)
	__gColW_Curve = (5, CSize.W, CSize.Wh, CSize.Wh)
	__gColW_Bone  = ((1,5), (2,CSize.W), (3,CSize.Wh-8), (4,CSize.W), (5,CSize.W), (6,1), (7,CSize.W),)
	__gColW_Morph = (5, CSize.W, CSize.Wh, CSize.Wh)

	#### =======================================================================
	def __init__(self, *arg, **kwd):
		self.wndName = self.WINDOWNAME
		self.ctls = BaseObject()
		self.subStore   = []
		self.presetPath = []
		self.tmpSelList = []
		self.lyrVisList = []
		self.dialogPath = ""

		(self.isDebug, self.isVerb, self.flags) = ParseLWO.DetermineVisOptions(**kwd)

	#### =======================================================================
	def SubWind_Reuse(self, wndName):
		if(mc.window(wndName, ex=True)):
			mc.showWindow(wndName)
			return(wndName, True)
		else:
			return(mc.window(wndName, t=wndName, s=True, rtf=True), False)

	#### ------------------------
	def SubWind_Rebuild(self, wndName):
		if(mc.window(wndName, ex=True)): mc.deleteUI(wndName)
		return(mc.window(wndName, t=wndName, s=True, rtf=True))

	#### =======================================================================
	def MenuSeparator(self, m, *arg, **kwd):
		mc.menuItem(p=m, d=True, **kwd)

	def MenuDelete(self, parent, *arg, **kwd):
		mc.popupMenu(parent, dai=True, e=True)	# 項目削除

	#### =======================================================================
	def SetCtrlHeight_ToSomeGroup(self, ctl, hgt):
		fullpath = mc.layout(ctl, fpn=True, q=True)
		chs = mc.layout(ctl, ca=True, q=True)
#		for c in chs:
#			ct = "%s|%s" % (fullpath, c)
#			mc.control(ct, h= hgt, e=True)
		mc.control("{}|{}".format(fullpath, chs[-1]), h= hgt, e=True)

	#### =======================================================================
	def SetupPopupMenu_LyrList(self, parent):
		menuCtl = mc.popupMenu(p= parent, b= PClick.MID)
		mc.menuItem(l= u"選択反転", c= partial(self.SelList_Mod_Rev, parent))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= u"全て選択", c= partial(self.SelList_Mod_AllSel, parent))
		mc.menuItem(l= u"全て解除", c= partial(self.SelList_Mod_DeSel, parent))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= u"選択: 可視",   c= partial(self.SelList_Mod_Vis,  parent, "sii"))
		mc.menuItem(l= u"選択: 不可視", c= partial(self.SelList_Mod_Hide, parent, "sii"))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= u"除外: 可視",   c= partial(self.SelList_Mod_Vis, parent, "dii"))
		mc.menuItem(l= u"除外: 不可視", c= partial(self.SelList_Mod_Hide,parent, "dii"))
		mc.setParent(m=True)	# 階層を閉じる

	#### ------------------------
	def SelList_Mod_Rev(self, ctl, *arg):
		u'''選択の反転'''
		n = mc.textScrollList(ctl, q=True, ni=True)		# 全項目数
		allLst = range(1, n+1)	# lst[0]=1,[1]=2...
		curSel = mc.textScrollList(ctl, q=True, sii=True)	# 選択操作のidxは 1 から

		if(not curSel):			# 選択無し
			mc.textScrollList(ctl, e=True, sii= allLst)
		else:
			# 選択が指す選択idxから配列idxへ-1,選択idxを0(無効値)へ.
			for i in curSel:
				allLst[(i -1)] = 0					# フラグとして 0

			reduced = list(set(allLst[:]))			# 重複削除,ソートされる.
			if(reduced[0] == 0): del reduced[0]		# 先頭に無効値(0)があれば除去.

			mc.textScrollList(ctl, e=True, da=True)	# 全選択解除
			if(len(reduced)): mc.textScrollList(ctl, e=True, sii= reduced)

	#### ------------------------
	def SelList_Mod_AllSel(self, ctl, *arg):
		u'''全て選択'''
		n = mc.textScrollList(ctl, q=True, ni=True)			# 全項目数
		mc.textScrollList(ctl, e=True, sii= range(1, n+1))	# 選択操作のidxは 1 から

	#### ------------------------
	def SelList_Mod_DeSel(self, ctl, *arg):
		u'''選択の全解除'''
		mc.textScrollList(ctl, e=True, da=True)

	#### ------------------------
	def SelList_Mod_Vis(self, ctl, opt, *arg):
		u'''可視レイヤを選択'''
		# 論理順で前詰めしたリストから可視属性を見る. 選択操作のidxは 1 から(start=1).
		visIdxList = []
		APD = visIdxList.append
		for (cnt, (i, isHide)) in enumerate(self.lyrVisList, start=1): # [論理idx,属性]
			if(not isHide):
				APD(cnt)
		if(visIdxList):
			mc.textScrollList(ctl, e=True, **{ opt: visIdxList })

	#### ------------------------
	def SelList_Mod_Hide(self, ctl, opt, *arg):
		u'''不可視レイヤを選択'''
		# 論理順で前詰めしたリストから可視属性を見る. 選択操作のidxは 1 から(start=1).
		hideIdxList = []
		APD = hideIdxList.append
		for (cnt, (i, isHide)) in enumerate(self.lyrVisList, start=1): # [論理idx,属性]
			if(isHide):
				APD(cnt)
		if(hideIdxList):
			mc.textScrollList(ctl, e=True, **{ opt: hideIdxList })

	#### =======================================================================
	def SetupPopupMenu_RecentName(self, parent, ctl):
		menuCtl = mc.popupMenu(p= parent, b= PClick.MID)
		mc.popupMenu(menuCtl, pmc= partial(self.MakePopupMenu_RecentName, menuCtl, ctl), e=True)

	#### ------------------------
	def MakePopupMenu_RecentName(self, parent, ctl, *arg):
		self.MenuDelete(parent)
		self.MakeSubMenu_Recent(parent, ctl)
		self.MakeSubMenu_PresetPath(parent, ctl)

	#### ------------------------
	def MakeSubMenu_Recent(self, parent, ctl):
		selID = mc.menuItem(p= parent, l= u"履歴", sm=True)
		if(len(self.subStore) == 0):
			mc.menuItem(selID, en=False, e=True)	#
		else:
			for n in self.subStore:
				mc.menuItem(p= selID, l= n, c= partial(self.Set_FileNameField_ByMenu, ctl, n))
		mc.setParent(m=True)	# 階層閉じ

	#### ------------------------
	def MakeSubMenu_PresetPath(self, parent, ctl):
		selID = mc.menuItem(p= parent, l= u"プリセット", sm=True)
		for n in self.presetPath:
			mc.menuItem(p= selID, l= n, c= partial(self.Set_FileNameField_ByMenu, ctl, n))
		mc.setParent(m=True)	# 階層閉じ

	#### ------------------------
	def RefineRecentNameArray(self, s):
		try:
			idx = self.subStore.index(s)
			del self.subStore[idx]
		except:
			pass
		self.subStore.insert(0, s)

	#### ------------------------
	@staticmethod
	def Determine_DialogPath(fn, *arg):
		u'''path or path/fileの有無チェック. 空の時はワークスペースRoot'''
		if(fn):
			if(os.path.exists(fn)):				# fileとして存在すればdirも存在する
				if(os.path.isfile(fn)):			# fileならdir抽出して返す
					pt = os.path.dirname(fn)
					if(os.path.exists(pt)): return(pt)
				else:
					return(fn)					# dirならそのまま
			else:
				return(fn)						# dirならそのまま
		return(mc.workspace(q=True, rd=True))	# 無効なpathならワークスペースRoot

	#### ------------------------
	def Set_FileNameField_ByMenu(self, ctl, filePathName, *arg):
		mc.textFieldButtonGrp(ctl, e=True, tx= filePathName)
		self.dialogPath = self.Determine_DialogPath(filePathName)
		self.Set_LayerList(filePathName)

	#### ------------------------
	def Get_InitialPath_FromArg(self, fn, *arg):
		u'''引数のfilePathの検証, 文字列としての評価, 文字列以外なら空文字, 配列なら[0]'''
		res = ""	# None,空文字,空配列
		if(fn):
			if(isinstance(fn, (list, tuple))):
				if(isinstance(fn[0], basestring)):	res = fn[0]
			elif(isinstance(fn, basestring)):		res = fn
			else: pass
		self.dialogPath = self.Determine_DialogPath(res)
		return(res)

	#### ------------------------
	@staticmethod
	def Get_PresetName_FromArg(*arg, **kwd):
		u'''引数のプリセットパスの検証'''
		if(kwd):
			p = kwd.get("p", kwd.get("preset", None))
			if(not p): return([])

			if(isinstance(p, basestring)): return([p])	# 単一文字列なら配列にして返す

			if(isinstance(p, (list, tuple))):	# 配列なら文字列だけをピックアップ
				return(filter(lambda x: isinstance(x, basestring), p))
		return([])

	#### =======================================================================
	def Create_Ctl_FileLine(self, parent, initialPath, *arg):
		ctl = mc.textFieldButtonGrp(p= parent, l=u"File", tx= initialPath, bl=u"　参照　", cw=((1,25),(3,50)), adj=2, \
			bc= partial(self.Get_FileName_FromDialog), cc= partial(self.Set_FileName_FromPanel))
		self.SetupPopupMenu_RecentName(ctl, ctl)
		self.SetCtrlHeight_ToSomeGroup(ctl, self.CSize.FHGT)
		mc.setParent(parent)
		return(ctl)

	#### =======================================================================
	def Create_Ctl_BaseLine(self, parent, *arg):
		ctls = self.ctls

		r1 = mc.rowLayout(p= parent, nc=5, cw5= self.__gColW_Base, adj=5)
		mc.text()
		mc.text(l=u"形状倍率")
		ctls.SCL = mc.floatField(v= 100.0)

		subForm = mc.formLayout(p= r1)
		ctls.SEL_LYR  = mc.iconTextCheckBox(l=u"[レイヤを選択]", ebg=True)
		ctls.IGN_HIDE = mc.checkBox(l=u"不可視を除外")
		mc.iconTextCheckBox(ctls.SEL_LYR, cc= partial(self.SyncStat_IgnoreHideLayer, ctls.SEL_LYR, ctls.IGN_HIDE), e=True)

		mc.button(p= r1, l=u"※ 読込み実行 ※", bgc= BGCols.MPOP_GREEN, c= partial(self.Exec))

		aco = ((ctls.IGN_HIDE, FPos.LT, 5, ctls.SEL_LYR),)	# ctl1の辺をctl2の辺に合わせる
		mc.formLayout(subForm, ac= aco, e=True)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Polygons(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=5, cw5= self.__gColW_PolMakes)
		mc.text()
		ctls.MESH = mc.iconTextCheckBox(l=u"[[ポリゴンを作成]]", v=True, cc= partial(self.SyncStat_Mesh))
		mc.iconTextCheckBox(ctls.MESH, cc= partial(self.SyncStat_Mesh, ctls.MESH), e=True)
		ctls.SHSG = mc.checkBox(l=u"重複SGを共用.",   v=True)
		ctls.UEDG = mc.checkBox(l=u"ｴｯｼﾞｳｪｲﾄを適用.", v=True)
		ctls.UPIV = mc.checkBox(l=u"Pivotを適用.")

		#### --------------------------------
		mc.rowLayout(p= parent, nc=5, cw5= self.__gColW_PolMakes)
		mc.text()
		ctls.VCOL = mc.checkBox(l=u"頂点色を適用.", v= True, cc= partial(self.SyncStat_VCol))
		ctls.MIVC = mc.checkBox(l=u"miVColorを適用.")
		ctls.DBLS = mc.checkBox(l=u"両面をON.")

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Sets(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=5, cw5= self.__gColW_Sets)
		mc.text()
		ctls.UTXT = mc.text(l=u"Setsへの登録:")
		ctls.PICK = mc.checkBox(l=u"頂点選択セット.")
		ctls.COLR = mc.checkBox(l=u"Colorタグ.")
		ctls.PART = mc.checkBox(l=u"Partsタグ.")

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Shader(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=3, cw3= self.__gColW_Shader)
		mc.text()
		ctls.STXT = mc.text(l=u"使用ｼｪｰﾀﾞ:")
		ctls.UBLN = mc.radioButtonGrp(sl= self.ShdType.LAMB, nrb=3, la3= self.__gShadLbls, cl3= self.__gShadAlign, cw3= self.__gShadWidth)
		# idxが1からなので+1

		#### --------------------------------
		mc.rowLayout(p= parent, nc=4, cw4= self.__gColW_Image)
		mc.text()
		ctls.TTXT = mc.text(l=u"画像を適用:")
		ctls.UCOL = mc.checkBox(l=u"Color")
		ctls.UBMP = mc.checkBox(l=u"Bump")

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Morphs(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=4, cw4= self.__gColW_Morph)
		mc.text()
		ctls.MTXT = mc.text(l=u"モーフ形状を作成 :")
		ctls.LREL = mc.checkBox(l=u"相対")
		ctls.LABS = mc.checkBox(l=u"絶対")

		#### --------------------------------
		mc.rowLayout(p= parent, nc=4, cw4= self.__gColW_Morph)
		mc.text()
		ctls.BTXT = mc.text(l=u"ﾌﾞﾚﾝﾄﾞｼｪｲﾌﾟを設定:")
		ctls.UREL = mc.checkBox(l=u"相対")
		ctls.UABS = mc.checkBox(l=u"絶対")

		mc.checkBox(ctls.LREL, e=True, cc= partial(self.SyncStat_Morph, ctls.LREL, ctls.UREL, ctls.LABS))
		mc.checkBox(ctls.LABS, e=True, cc= partial(self.SyncStat_Morph, ctls.LABS, ctls.UABS, ctls.LREL))

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Bones(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=7, cw= self.__gColW_Bone, cal=(6,"right"), adj=6)
		mc.text()
		ctls.BONE = mc.iconTextCheckBox(l=u"[[ボ－ンを作成]]")
		mc.iconTextCheckBox(ctls.BONE, cc= partial(self.SyncStat_Bone, ctls.BONE), e=True)

		ctls.ZAXZ = mc.iconTextCheckBox(l=u"[ZAxis]", v=True)
		mc.iconTextCheckBox(ctls.ZAXZ, cc= partial(self.SyncStat_Wght, ctls.ZAXZ), e=True)

		ctls.WXML = mc.iconTextCheckBox(l=u"[ｳｪｲﾄのXML出力]")
		mc.iconTextCheckBox(ctls.WXML, cc= partial(self.SyncStat_Wght, ctls.WXML), e=True)

		ctls.UWGT = mc.iconTextCheckBox(l=u"[ｳｪｲﾄをAttr付与]")
		mc.iconTextCheckBox(ctls.UWGT, cc= partial(self.SyncStat_Wght, ctls.UWGT), e=True)

		mc.text()

		ctl = mc.button(l=u"▽ウェイト操作")
		menuCtl = mc.popupMenu(p= ctl, b= PClick.LFT)
		mc.menuItem(p= menuCtl, l= u"Attrから転送", c= partial(ApplyWeightsByAttr.Apply))
		mc.menuItem(p= menuCtl, l= u"Attrを除去",   c= partial(ApplyWeightsByAttr.Remove))

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Curves(self, parent, *arg):
		ctls = self.ctls

		mc.rowLayout(p= parent, nc=4, cw4= self.__gColW_Curve)
		mc.text()
		ctls.CURV = mc.iconTextCheckBox(l=u"[[カーブを作成]]")
		mc.iconTextCheckBox(ctls.CURV, cc= partial(self.SyncStat_Crv, ctls.CURV), e=True)
		ctls.DTXT = mc.text(l=u"ｶｰﾌﾞ次数")
		ctls.DEG  = mc.intField(min=1, max=7, v=3)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_LayerList1(self, parent, baseClm, *arg):
		u'''レイヤーリスト作成1'''
		ctls = self.ctls

#		subClm = mc.colmnLayout(p= rootForm, adj=True, bgc=(0,1,0))
		subClm = mc.rowLayout(p= parent, nc=1, rat=(1,"both",0), adj=1)#, bgc=(0,1,0))
		afo = (
			(baseClm, FPos.LT, 0), (baseClm, FPos.RT, 0), (baseClm, FPos.TP, 0),
			(subClm,  FPos.LT, 0), (subClm,  FPos.RT, 0),#  (subClm, FPos.BT, 0),
		)
		aco = ((subClm, FPos.TP, 0, baseClm),)
		apo = ((subClm, FPos.BT, 0, 100),)
		mc.formLayout(parent, e=True, af= afo, ac= aco, ap=apo)

		lstForm = mc.formLayout(p= subClm)

		lstCtl = mc.textScrollList(p= lstForm, fn= LblFont.FIX, ams=True)

#		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0),)
#		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0), (lstCtl, FPos.BT, 0),)
		apo = ((lstCtl, FPos.BT, 0, 100),)
		afo =()
		mc.formLayout(lstForm, e=True, af= afo, ap= apo)

		return(lstCtl)

	#### =======================================================================
	def Create_Ctl_LayerList2(self, parent, baseClm, *arg):
		u'''レイヤーリスト作成2'''
		ctls = self.ctls

		lstForm = mc.formLayout(p= parent)
		afo = (
			(baseClm, FPos.LT, 0), (baseClm, FPos.RT, 0), (baseClm, FPos.TP, 0),
			(lstForm, FPos.LT, 0), (lstForm, FPos.RT, 0), (lstForm, FPos.BT, 0),
		)
		aco = ((lstForm, FPos.TP, 0, baseClm),)
		mc.formLayout(parent, e=True, af= afo, ac= aco)

		lstCtl = mc.textScrollList(p= lstForm, fn= LblFont.FIX, ams=True)
		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0),)
		apo = ((lstCtl, FPos.BT, 0, 100),)
		mc.formLayout(lstForm, e=True, af= afo, ap= apo)

		return(lstCtl)

	#### =======================================================================
	def SetupUITemplate_Base(self, name):
		u'''UIテンプレート'''
		CSize = self.CSize

		if(mc.uiTemplate(name, ex=True)): mc.deleteUI(name, uit=True)
		mc.uiTemplate(name)

		mc.rowLayout(dt=name, h= CSize.RHGT)
		mc.button(dt=name, h= CSize.FHGT)
		mc.intField(dt=name, h= CSize.FHGT, w= CSize.CRV, s=1)
		mc.floatField(dt=name, h= CSize.FHGT, w= CSize.SCL, pre=2, s=1.0, min=0.01, max=10000.0)
		mc.text(dt=name, l=u"", h= CSize.FHGT)
		mc.textFieldButtonGrp(dt=name, h= CSize.RHGT)
		mc.iconTextCheckBox(dt=name, h= CSize.FHGT, st= ICStyle.TEXT, ebg=True, v=False)
		mc.checkBox(dt=name, h= CSize.FHGT, v=False)

	#### =======================================================================
	def Window(self, fn, *arg, **kwd):
#		(self.wndName, isReopen) = self.SubWind_Reuse(self.wndName)
#		if(isReopen): return
		self.wndName = self.SubWind_Rebuild(self.wndName)

		self.presetPath = self.Get_PresetName_FromArg(**kwd)
		ctls = self.ctls

		self.SetupUITemplate_Base(self.TEMPLA_BASE)		# make Template

		mc.setUITemplate(self.TEMPLA_BASE, pst=True)	# push Template

		rootForm = mc.formLayout()
		baseClm  = mc.columnLayout(p=rootForm, adj=True)

		#### --------------------------------
		initialPath = self.Get_InitialPath_FromArg(fn)
		ctls.FNAME = self.Create_Ctl_FileLine(baseClm, initialPath)	# ファイル

		self.Create_Ctl_BaseLine(baseClm)	# 読込み
		mc.separator()

		self.Create_Ctl_Polygons(baseClm)	# ポリ
		mc.separator()

		self.Create_Ctl_Sets(baseClm)		# セット
		mc.separator()

		self.Create_Ctl_Shader(baseClm)		# シェーダ
		mc.separator()

		self.Create_Ctl_Morphs(baseClm)		# モーフ
		mc.separator()

		self.Create_Ctl_Bones(baseClm)		# ボーン
		mc.separator()

		self.Create_Ctl_Curves(baseClm)		# カーブ
		mc.separator()

		#### レイヤリスト --------------------------------
		if(False):
			ctls.LYRS = self.Create_Ctl_LayerList1(rootForm, baseClm)
		else:
			ctls.LYRS = self.Create_Ctl_LayerList2(rootForm, baseClm)
			#(こっちが有効)
		self.SetupPopupMenu_LyrList(ctls.LYRS)

		# ステータス反映
		self.SyncStat_Mesh(ctls.MESH)
		self.SyncStat_VCol()
		self.SyncStat_Morph(ctls.LREL, ctls.UREL, ctls.LABS)
		self.SyncStat_Morph(ctls.LABS, ctls.UABS, ctls.LREL)
		self.SyncStat_Bone(ctls.BONE)
		self.SyncStat_Wght(ctls.ZAXZ)
		self.SyncStat_Wght(ctls.UWGT)
		self.SyncStat_Wght(ctls.WXML)
		self.SyncStat_Crv(ctls.CURV)
		self.SyncStat_IgnoreHideLayer(ctls.SEL_LYR, ctls.IGN_HIDE)

#		lstC = mc.columnLayout(adj=True)
#		afo = ((lstC, FPos.TP, 0), (lstC, FPos.LT, 0), (lstC, FPos.RT, 0),)
#		apo = ((lstC, FPos.BT, 0, 100),)
		self.Set_LayerList(initialPath)	# ID_LYRS作成後に処理

		mc.setUITemplate(ppt=True)	# pop Template(BASE)

		mc.showWindow(self.wndName)

	#### =======================================================================
	@staticmethod
	def EnableControl(ctl, stat):
		mc.control(ctl, en= stat, e=True)

	#### -----------------------------------------------------------------------
	def ChangeButtonCol(self, ctl, *arg):
		u'''ボタン色変更'''
		isAct = mc.iconTextCheckBox(ctl, q=True, en=True)
		isEbl = mc.iconTextCheckBox(ctl, q=True, v=True)
		BGC = BGCols.LPOP_BLUE if(isAct and isEbl) else BGCols.BASE3
		mc.iconTextCheckBox(ctl, e=True, bgc= BGC)

	#### --------------------------------
	def SyncStat_Crv(self, ctl, *arg):
		u'''カーブを作成に反応'''
		self.ChangeButtonCol(ctl)
		stat = mc.iconTextCheckBox(ctl, q=True, v=True)
		self.EnableControl(self.ctls.DEG,  stat)
		self.EnableControl(self.ctls.DTXT, stat)

	#### --------------------------------
	def SyncStat_Bone(self, ctl, *arg):
		u'''ボーンを作成に反応'''
		self.ChangeButtonCol(ctl)
		stat = mc.iconTextCheckBox(ctl, q=True, v=True)
		self.EnableControl(self.ctls.ZAXZ,  stat)
		self.SyncStat_Wght(self.ctls.ZAXZ)

	#### --------------------------------
	def SyncStat_Wght(self, ctl, *arg):
		u'''ウェイトを読込に反応'''
		self.ChangeButtonCol(ctl)

	#### --------------------------------
	def SyncStat_Mesh(self, ctl, *arg):
		u'''ポリゴンを作成に反応'''
		self.ChangeButtonCol(ctl)
		stat = mc.iconTextCheckBox(ctl, q=True, v=True)
		ctls = self.ctls
		tgt = (
			ctls.UEDG,ctls.SHSG,ctls.TTXT,ctls.VCOL,ctls.MIVC,ctls.UTXT,ctls.UPIV,
			ctls.UCOL,ctls.UBMP,ctls.PICK,ctls.COLR,ctls.PART,ctls.STXT,ctls.DBLS,ctls.UBLN,
			ctls.WXML,ctls.UWGT,
		)
		for ctl in tgt:
			self.EnableControl(ctl, stat)

	#### --------------------------------
	def SyncStat_IgnoreHideLayer(self, ctlSelLyr, ctlIgnore, *arg):
		u'''レイヤ選択/不可視を除外'''
		self.ChangeButtonCol(ctlSelLyr)
		stat = mc.iconTextCheckBox(ctlSelLyr, q=True, v=True)
		self.EnableControl(ctlIgnore, stat ^1)

		self.EnableControl(self.ctls.LYRS, stat)
		if(stat):
			mc.textScrollList(self.ctls.LYRS, e=True, si= self.tmpSelList)	# 選択
		else:
			self.tmpSelList = mc.textScrollList(self.ctls.LYRS, q=True, si=True) or []
			mc.textScrollList(self.ctls.LYRS, e=True, da=True)			# 選択クリア

	#### --------------------------------
	def SyncStat_Morph(self, ctlMake, ctlSet, ctlSubSet, *arg):
		u'''モーフを作成に反応'''
		stat = mc.checkBox(ctlMake, q=True, v=True)
		self.EnableControl(ctlSet, stat)

		stat2 = mc.checkBox(ctlSubSet, q=True, v=True)
		self.EnableControl(self.ctls.BTXT, (stat|stat2))

	#### --------------------------------
	def SyncStat_VCol(self, *arg):
		u'''頂点色を適用に反応'''
		self.EnableControl(self.ctls.MIVC, mc.checkBox(self.ctls.VCOL, q=True, v=True))

	#### =======================================================================
	def Get_FileName_FromDialog(self, *arg):
		dialogPref = mc.optionVar(q= "FileDialogStyle")

		fn = mc.fileDialog2(cap= self.FILE_CAPTION, dir= self.dialogPath, ff= self.FILE_FILTER, ds= dialogPref, fm=1)
		if(not fn): return	# キャンセルはNone, 戻りは[u"drv:/path/file", ...], fm=4時はds=1/2でセパレータが変わる

		self.dialogPath = mc.workspace(q=True, dir=True)	# 前回作業Dirとして

		mc.textFieldButtonGrp(self.ctls.FNAME, e=True, tx= fn[0])

		self.RefineRecentNameArray(fn[0])	# ファイル履歴
		self.Set_LayerList(fn[0])			# lyr情報

	#### --------------------------------
	def Set_FileName_FromPanel(self, *arg):
		filePathName = mc.textFieldButtonGrp(self.ctls.FNAME, q=True, tx=True)
		(base, ext) = os.path.splitext(filePathName)

		if(ext.lower() != self.FILE_EXT):	# 拡張子が無ければ補完, ※補完後の有無をチェックしてない
			filePathName += self.FILE_EXT
			mc.textFieldButtonGrp(self.ctls.FNAME, e=True, tx= filePathName)

		self.dialogPath = self.Determine_DialogPath(filePathName)
		self.Set_LayerList(filePathName)

	#### =======================================================================
	def Set_LayerList(self, filename=None, *arg):
		# 起動直後、ダイアログ、パス入力時に呼ばれる
		if(not filename): return

		# layersは格納順/物理順. lyr.idxは論理順(ModelerGUI).
		layers = ParseLWO.PreScan_LWO(filename.encode())		# 戻りはlwo内格納順
		if(layers == None): return

		maxNameLen = 0
		for l in layers:
			maxNameLen = max(maxNameLen, len(l.name))
		maxNameLen += 1

		(lyrVisList, layerList) = ([], [])
		visAPD = lyrVisList.append
		lyrAPD = layerList.append
		FMT_BASE = ": {:<{pad}s}:VC/{} Edg/{} Rel/{} Abs/{} Bon/{} Crv/{}"
		FMT_VIS  = "{:03}({:03})" + FMT_BASE
		FMT_HIDE = "{:03}[{:03}]" + FMT_BASE
		mk = ("x", "o")
		for (fi, lyr) in enumerate(layers):	# 物理順で列挙
			hasVC  = mk[bool(lyr.rgb)|bool(lyr.rgba)]
			hasEdg = mk[bool(lyr.hasEdge|lyr.hasEdgeM)]
			hasRel = mk[bool(lyr.relMorphs)]
			hasAbs = mk[bool(lyr.absMorphs)]
			hasBon = mk[bool(lyr.bonNum)]
			hasCrv = mk[bool(lyr.crvNum)]
			pad = maxNameLen - len(lyr.name)

			# 表示上,ModelerのGUI合わせでlyr.idx +1する
			FMT = FMT_HIDE if(lyr.isHide) else FMT_VIS
			lineStr = FMT.format(lyr.idx +1, fi, lyr.name, hasVC, hasEdg, hasRel, hasAbs, hasBon, hasCrv, pad=maxNameLen)
			lyrAPD(lineStr)
			visAPD([lyr.idx, lyr.isHide])	# [論理,属性]

		# 論理順でソートして格納
		self.lyrVisList = sorted(lyrVisList, key= lambda x: x[0])
		layerList.sort()

		ctls = self.ctls
		mc.textScrollList(ctls.LYRS, e=True, ra=True, a= layerList)	# リストクリア, 項目追加は同時可

		stat = mc.iconTextCheckBox(ctls.SEL_LYR, q=True, v=True)
		if(stat):
			mc.textScrollList(ctls.LYRS, e=True, si= layerList)		# 選択はクリア/項目追加と同時不可

		self.tmpSelList = layerList[:]	# 新規項目の退避

	#### =======================================================================
	@staticmethod
	def Get_Options(ctls= None, *arg):
		opts = BuildLWO.OptsObject()
		if(not ctls): return(opts)

		QV = {"q":True, "v":True}

		for (o, c) in (("isMesh","MESH"), ("isBone","BONE"), ("isCrv","CURV"), ("useWgt","UWGT"), ("useXML","WXML"), ("isZAxis", "ZAXZ")):
			opts[o] = mc.iconTextCheckBox(ctls[c], **QV)

		arCBoxs = (
			("useEdge",  "UEDG"), ("usePivot","UPIV"),
			("loadRM",   "LREL"), ("loadAM",  "LABS"), ("useRM", "UREL"), ("useAM", "UABS"),
			("useVCol",  "VCOL"), ("useMiVC", "MIVC"),
			("shareSG",  "SHSG"), ("useColT", "UCOL"), ("useBmpT", "UBMP"),
			("usePick",  "PICK"), ("useColr", "COLR"), ("usePart", "PART"),
			("isDblSide","DBLS"), ("skipHide","IGN_HIDE")
		)
		for (o, c) in arCBoxs:
			opts[o] = mc.checkBox(ctls[c], **QV)
		opts.scl    = mc.floatField(ctls.SCL, **QV)
		opts.deg    = mc.intField(  ctls.DEG, **QV)
		opts.mtlType= mc.radioButtonGrp(ctls.UBLN, q=True, sl=True) -1	# 戻りは1からなので-1
		opts.vScl = [opts.scl, opts.scl, -opts.scl]
		return(opts)

	#### =======================================================================
	def Exec(self, *arg, **kwd):
		ctls = self.ctls

		fname= mc.textFieldButtonGrp(ctls.FNAME, q=True, tx=True)
		if(not os.path.isfile(fname)):
			mc.warning(u"ファイル選択が無い."),
			return

		# actListはfile内格納順に応じたboolのリスト
		actList = None

		useLyrSel = mc.iconTextCheckBox(ctls.SEL_LYR, q=True, v=True)
		if(useLyrSel):
			selList = mc.textScrollList(ctls.LYRS, q=True, si=True) or []	# GUI上の項目名のリスト
			if(not selList):
				mc.warning(u"レイヤー選択が無い."),
				return

			allNum  = mc.textScrollList(ctls.LYRS, q=True, ni=True)	# 時点のGUI上のリスト総数
			actList = [False] * allNum

			# GUIIdx1(FileIdx0): name~ | GUIIdx1[FileIdx0]: name~, 可視属性によって()[]が違う
			pat = re.compile("^\d+[\(\[](\d+)[\)\]]:")
			for s in selList:
				idx = int(pat.findall(s)[0])	# 戻りはリスト
				actList[idx] = True

		opts = self.Get_Options(self.ctls)
		if((opts.isMesh|opts.isBone|opts.isCrv) == 0):
			mc.warning(u"読み込み対象が未指定."),
			return

		(tags, layers, surfs, clips) = ParseLWO.Open_LWO(fname.encode(), d=self.isDebug, v=self.isVerb, f=self.flags)
		if(not tags):
			return
		BuildLWO.BuildObjects(tags, layers, surfs, clips, o=opts, al=actList)

		self.RefineRecentNameArray(fname)	# ファイル履歴
