﻿# -*- coding: utf-8 -*-

#### =======================================================================
__all__ = ("OptsObject",)

#### ========================================================================
class OptsObject(object):
	__slots__ = (
		"isMesh", "isBone", "isCrv", "scl", "vScl", "deg", "useEdge", "usePivot",
		"loadRM", "loadAM", "useRM", "useAM", "useWgt", "useXML", "isZAxis", "isDblSide",
		"useVCol", "useMiVC", "shareSG", "useColT", "useBmpT",
		"usePick", "useColr", "usePart",
		"mtlType", "skipHide",
	)
	DEFAULT_SCL = 100.0
	DEFAULT_DEG = 3
	(LAMB, BLIN, PHON, SURFS) = (0,1,2,3)	# build用故0から
	#### -----------------------------------------------------------------------
	def __init__(self, **kwd):
		scl = kwd.get("s", self.DEFAULT_SCL)
		deg = kwd.get("d", self.DEFAULT_DEG)
		mtl = kwd.get("m", self.LAMB)
		self.isMesh  = True		# Mesh作成
		self.isBone  = False	# Bone作成
		self.isCrv   = False	# Curve作成
		self.scl     = scl		# 倍率(UI用)
		self.deg     = deg		# Curve次数
		self.useEdge = True		# エッジウェイト適用
		self.usePivot= False	# ピボット適用
		self.loadRM  = False	# モーフ(相対)形状作成
		self.loadAM  = False	# モーフ(絶対)形状作成
		self.useRM   = False	# モーフ(相対)をBlendShapeNodeに適用
		self.useAM   = False	# モーフ(絶対)をBlendShapeNodeに適用
		self.shareSG = True		# 既存の同名SGを共用
		self.useVCol = False	# 頂点カラーのmeshNodeへの適用
		self.useMiVC = False	# mentalrayVertexColorsNodeの適用
		self.usePick = False	# 頂点選択セット(PICK)をSetsに適用
		self.useColr = False	# 面カラー(COLR)をSetsに適用
		self.usePart = False	# 面パート(PART)をSetsに適用
		self.useColT = False	# Colorの画像からfileNode+layeredTextureを作成/接続
		self.useBmpT = False	# Bumpの画像からfileNode+layeredTexture+Bump2dを作成/接続
		self.mtlType = mtl		# 使用マテリアルタイプ(0:lambert, 1:blinn, 2:phong, x:SurfaceShader)
		self.skipHide= False	# 不可視属性Layerを無視
		self.useWgt  = False	# (Layer内にBoneがある場合)ウェイト値をMesh|Shape.attrに書込み
		self.useXML  = False	# (Layer内にBoneがある場合)ウェイト値をdeformerWeight用XMLを書き出し
		self.isZAxis = True		# LWスタイルのボーンとして扱う
		self.isDblSide=False	# Shapeオプション: 両面
		self.vScl    = [scl,scl,-scl]	# Z座標反転した倍率ベクトル(Build用)
	#### ----------------
	def items(self):
		return([ (k, getattr(self, k)) for k in self.__slots__ ])
	def values(self):
		return([ getattr(self, k) for k in self.__slots__ ])
	def keys(self):
		return([ k for k in self.__slots__ ])
	def __getitem__(self, s):
		return(getattr(self, s))
	def __setitem__(self, s, v):
		setattr(self, s, v)
