﻿# -*- coding: utf-8 -*-

import maya.OpenMaya as om
import maya.cmds as mc

import FnOM1
from ..LWO_ParseHeader import *
from .Defines import *

#### ========================================================================
__all__ = ("MakeSurface", "SetSurfaceAttr",)

#### ========================================================================
__DEBUG = 0

#### ========================================================================
def __MakeSurf_GetInitialMaterialOutPlug():
	u'''初期lambertのocプラグを取るだけ'''
	fn = om.MFnDependencyNode(FnOM1.GetDependNode(FnOM1.MakeSelectionList("lambert1"), 0))
	return(fn.findPlug("oc"))

#### ----------------
def __MakeSurf_DetermineShaderNode(type):
	u'''作成Mtlの型決定'''
	if(type == ShdType.BLIN):
		return(om.MFnBlinnShader())
	elif(type == ShdType.PHON):
		return(om.MFnPhongShader())
	elif(type == ShdType.SURFS):
		return(om.MFnLambertShader())	# 仮

	return(om.MFnLambertShader())	# lambert

#### ----------------
def __MakeSurf_AddMember_ToExsitingSG(sgName, selLst):
	u'''既存のSGにセット'''
	Fn = om.MFnSet(FnOM1.GetDependNode(FnOM1.MakeSelectionList(sgName), 0))
	Fn.addMembers(selLst)

#### ----------------
def __MakeSurf_DetermineTargetList(dagpMesh, surfNum, polIdxs):
	u'''コンポーネント選択リスト作成'''
	FnMe = om.MFnMesh(dagpMesh)
	selLst = om.MSelectionList()

	# surfが1つ、pol数一致ならcompでなくdagへの適用
	if((surfNum == 1) and (len(polIdxs) == FnMe.numPolygons())):
		selLst.add(dagpMesh)
	else:
		Fn = om.MFnSingleIndexedComponent()
		comp = Fn.create(om.MFn.kMeshPolygonComponent)
		Fn.addElements(FnOM1.MIntArrayFromList(polIdxs))
		selLst.add(dagpMesh, comp)		# コンポーネント選択リスト作成
	return(selLst)

#### ----------------
def MakeSurface(tags, surfTags, buildedPolIdxs, dagpMesh, opts):
	if(__DEBUG): print("MakeSurface:")

	plgIniOut = __MakeSurf_GetInitialMaterialOutPlug()	# 初期マテリアル outColor

	FnShd = __MakeSurf_DetermineShaderNode(opts.mtlType)	# 作成Mtl型

	FnSet= om.MFnSet()
	MDGM = om.MDGModifier()

	surfNum = len(surfTags.keys())

	(mtNames, sgNames) = ({}, {})
	for i in surfTags.keys():
		mtName = tags[i]
		if(mtName.rfind(SFixNM.MTL) == -1):	# _MTが無い(_MT6等,中間出現を許容)
			mtName = "{}{}".format(mtName, SFixNM.MTL)
		sgName = mtName.replace(SFixNM.MTL, SFixNM.SG)	# 改名後に合わせてreplacce

		# コンポーネント選択リスト作成
		polIdxs = tuple( buildedPolIdxs[polIdx] for polIdx in surfTags[i] if(buildedPolIdxs[polIdx]!=None) )
		selLst = __MakeSurf_DetermineTargetList(dagpMesh, surfNum, polIdxs)

		if(opts.shareSG and mc.objExists(sgName)):	# 既存SGが有れば利用して次へ
			__MakeSurf_AddMember_ToExsitingSG(sgName, selLst)
			continue

		FnSet.create(selLst, FnSet.kRenderableOnly)	# SG作成. 初期マテリアルが繋がってる
		# result強制表示2012~15, 消せない.

		FnShd.create()	# Mtl作成

		plgInp = FnSet.findPlug("ss")		# surfaceShader
		MDGM.disconnect(plgIniOut, plgInp)	# 初期マテリアル切断
		MDGM.connect(FnShd.findPlug("oc"), plgInp)
		MDGM.doIt()	# MDGModifierはdoIt()で処理確定

		mtName = FnShd.setName(mtName)		# 重複改名があるので先行
		sgName = FnSet.setName(sgName)
		mtNames[i] = [tags[i], mtName]
		sgNames[i] = sgName
	return(mtNames, sgNames)

#### ------------------------------------------------------------------------
def __BuildTextureMap(imaps, clips, newName, useColT, useBmpT):
	# 現状有効なマップタイプをピックアップ
	activeMaps = filter(lambda mp: (mp.project == TXPRJ.UVMAP), imaps)
	activeMaps.sort(key= lambda x: x.ord)	# 序数でソート

	(colNum, bmpNum) = (0, 0)
	for mp in activeMaps:					# レイヤ用、現状で有効な属性毎のレイヤ数
		if(  mp.chan == SurfT.COLR): colNum += 1
		elif(mp.chan == SurfT.BUMP): bmpNum += 1

	(colLtName, bumpLtName) = ("", "")
	if(colNum * useColT):
		colLtName  = mc.shadingNode(NdType.LAYR, n= newName.replace(SFixNM.MTL, SFixNM.LWCOL), at=True)
		mc.connectAttr("{}.oc".format(colLtName), "{}.c".format(newName), f=True)

	if(bmpNum * useBmpT):
		bumpLtName = mc.shadingNode(NdType.LAYR,  n= newName.replace(SFixNM.MTL, SFixNM.LWBMP), at=True)
		bmpName = mc.shadingNode(NdType.BUMP, n= NameFMT.B2D.format(NAME= newName.replace(SFixNM.MTL, "")), at=True)
		mc.setAttr("{}.ail".format(bumpLtName), True)								# alphaIsLuminance
		mc.connectAttr("{}.oa".format(bumpLtName), "{}.bv".format(bmpName), f=True)	# outAlpha->bumpValue
		mc.connectAttr("{}.o".format(bmpName), "{}.n".format(newName), f=True)		# outNormala->normalCamera

	for mp in activeMaps:
		if(mp.imgIdx == 0):	# ファイル未指定時は0、clips[0]は存在しない
			fpath = ""
		else:
			fpath = clips[mp.imgIdx].stillName
			if((fpath[1] == ":") and (fpath[2] != "\\")):
				fpath = fpath[0:2] + "\\" + fpath[2:]

		if(fpath != ""):
			r = fpath.rsplit("/", 1)
			r1 = r[0] if(len(r) < 2) else r[1]
			ndName = r1.split(".")[0] + SFixNM.MTL
		else:
			ndName = newName
		name = __Make_FileNode(ndName)

		if(fpath != ""):
			mc.setAttr("{}.ftn".format(name), fpath, **VType.STR)
		if(useColT and (mp.chan == SurfT.COLR)):
			idx = mc.getAttr("{}.cs".format(colLtName), s=True)
			mc.connectAttr("{}.oc".format(name), "{}.cs[{}].c".format(colLtName, idx), f=True)
		if(useBmpT and (mp.chan == SurfT.BUMP)):
			idx = mc.getAttr("{}.cs".format(bumpLtName), s=True)
			mc.connectAttr("{}.oc".format(name), "{}.cs[{}].c".format(bumpLtName, idx), f=True)

#### ------------------------------------------------------------------------
def __ConnectVCol(vcolName, mapNames, newName):
	if(not mapNames): return	# surfaceでVcol指定をしていてもmeshにVcolが無い事がある
	if(vcolName in mapNames):
		mc.connectAttr("{}.oc".format(NameFMT.VCOL.format(NAME= mapNames[vcolName])), "{}.c".format(newName), f=True)	# outColor, color

#### ------------------------------------------------------------------------
def SetSurfaceAttr(surfs, clips, mtNames, sgNames, rgbNames, rgbaNames, opts):
	if(__DEBUG): print("SetSurfaceAttr:")

	for i in mtNames.keys():
		(oldName, newName) = (mtNames[i][0], mtNames[i][1])
		srf = surfs[oldName]
		mc.setAttr("{}.c".format(newName),    *srf.colr, **VType.DBL3)			# color
		mc.setAttr("{}.dc".format(newName),    srf.diff)						# diffuse
		mc.setAttr("{}.ambc".format(newName), *([srf.lumi]*3), **VType.DBL3)	# ambientColor
		mc.setAttr("{}.it".format(newName),   *([srf.tran]*3), **VType.DBL3)	# transparency
		mc.setAttr("{}.tc".format(newName),    srf.trnl)						# translucence
		if(opts.useMiVC and (len(srf.vcol) != 0)):
			mapNames = rgbNames if(srf.vcol[1] == VMapT.RGB_) else rgbaNames
			__ConnectVCol(srf.vcol[0], mapNames, newName)
		if((opts.useColT|opts.useBmpT) and len(srf.imap)):
			__BuildTextureMap(srf.imap, clips, newName, opts.useColT, opts.useBmpT)

#### ========================================================================
def __Make_FileNode(ndName):
	# coverage, translateFrame, rotateFrame, mirrorU, mirrorV, stagger, wrapU, wrapV, repeatUV, offset, rotateUV, noiseUV,
	# vertexUvOne, vertexUvTwo, vertexUvThree, vertexCameraOne, outUV/uvCoord, out/uvFilterSize
	p2At = (".c",".tf",".rf",".mu",".mv",".s",".wu",".wv",".re",".of",".r", ".n",".vt1",".vt2",".vt3",".vc1",".o", ".ofs",)
	fnAt = (".c",".tf",".rf",".mu",".mv",".s",".wu",".wv",".re",".of",".ro",".n",".vt1",".vt2",".vt3",".vc1",".uv",".fs",)

	bsName = ndName.replace(SFixNM.MTL, SFixNM.FILE)
	fName  = mc.shadingNode(NdType.FILE, n= bsName, at=True)
	p2Name = mc.shadingNode(NdType.P2D,  n= NameFMT.P2D.format(NAME= bsName), au=True)
	mc.setAttr("{}.ft".format(fName), 0)		# filterType OFF
	for (pa,fa) in zip(p2At, fnAt):
		mc.connectAttr(p2Name + pa, fName + fa, f=True)
	return(fName)
