﻿# -*- coding: utf-8 -*-

# print()の動作をv3スタイルに.
from __future__ import print_function

import sys, os, chunk, struct

if(0):
	if(sys.version_info[0] == 2):
		import LWO_ParseHeader, LWO_Object
		reload(LWO_ParseHeader)
		reload(LWO_Object)
	else:
		import importlib
		from . import LWO_ParseHeader, LWO_Object
		importlib.reload(LWO_ParseHeader)
		importlib.reload(LWO_Object)
from .LWO_ParseHeader import *
from .LWO_Object import *

#### =======================================================================
__all__ = ("DetermineVisOptions", "PreScan_LWO", "Open_LWO")

#### ========================================================================
class VisFlags(object):
	u'''flag:
0x00001:PNT,  0x00002:POL,  0x00004:CRV,  0x00008:BONE
0x00010:UV,   0x00020:WGHT, 0x00040:MORF, 0x00080:VCOL
0x00100:EDGE, 0x00200:NORM, 0x00400:SELS, 0x00800:DMY1
0x01000:SURF, 0x02000:IMAP, 0x04000:CLIP, 0x08000:PSURF
0x10000:LAYR, 0x20000:TAGS, 0x40000:PTAG'''
	__slots__ = (
		"PNT",  "POL",  "CRV",  "BONE",
		"UV",   "WGHT", "MORF", "VCOL",
		"EDGE", "NORM", "SELS", "DMY1",
		"SURF", "IMAP", "CLIP", "PSURF",
		"LAYR", "TAGS", "PTAG",
	)
	__FMAX = len(__slots__)

	(F_PNT,  F_POL,  F_CRV,  F_BONE) = (0x00001, 0x00002, 0x00004, 0x00008)
	(F_UV,   F_WGHT, F_MORF, F_VCOL) = (0x00010, 0x00020, 0x00040, 0x00080)
	(F_EDGE, F_NORM, F_SELS, F_DMY1) = (0x00100, 0x00200, 0x00400, 0x00800)
	(F_SURF, F_IMAP, F_CLIP, F_PSURF)= (0x01000, 0x02000, 0x04000, 0x08000)
	(F_LAYR, F_TAGS, F_PTAG, F_DMY2) = (0x10000, 0x20000, 0x40000, 0x80000)
	F_ALL = 0x8FFFF

	#### -----------------------------------------------------------------------
	def __init__(self, flags=0, *arg):
		f = 0 if(flags == None) or (not isinstance(flags, int)) else flags
		for (i, s) in enumerate(self.__slots__[:self.__FMAX]):
			setattr(self, s, ((f>>i)&0x1))

	#### -----------------------------------------------------------------------
	def setFlags(self, flags=0, *arg):
		f = 0 if(flags == None) or (not isinstance(flags, (int, bool))) else flags
		for (i, s) in enumerate(self.__slots__[:self.__FMAX]):
			setattr(self, s, ((f>>i)&0x1))

	#### -----------------------------------------------------------------------
	def getFlags(self, *arg):
		f = 0
		for (i, v) in enumerate(self.__slots__[:self.__FMAX]):
			f |= (getattr(self, v) << i)
		return(f)

#### ========================================================================
#### ========================================================================
def DetermineVisOptions(**kwd):
	u'''引数の取得, 不適切な場合はデフォ値へ変更.'''
	__GetOpt = \
		(lambda kwd,sn,ln,dv:(lambda v:v if(isinstance(v,(int,bool))) else dv)(kwd.get(sn,kwd.get(ln,dv))))
	dbg  = __GetOpt(kwd, "d", "dbg", False)
	verb = __GetOpt(kwd, "v", "verbose", False)
	flag = __GetOpt(kwd, "f", "flag", 0)
	return(dbg, verb, flag)

#### ========================================================================
def __Check_FileState(fileName):
	u'''ファイルチェック'''
	if(not fileName) or (not isinstance(fileName, str)):
		return(None, u"ファイルが未指定.\n")

	if(not os.path.exists(fileName)):	# filenameを含まないpathも通る
		return(None, u"ファイル/パスが無い.\n")

	if(not os.path.isfile(fileName)):	# pathがfilenameを含むか
		return(None, u"ファイルが無い.\n")

	fp = open(fileName, 'rb')
	try:
		(tag, size, id) = struct.unpack(ReadT.FORM, fp.read(RSize.FORM))	# FORM:size:LWO2
	except:
		fp.close()
		return(None, u"ファイルがおかしい.\n")

	if(id not in { ChunkT.LWO2, ChunkT.LWO3 }):
		fp.close()
		return(None, u"LWO2/3以外は非サポート.\n")

	return(fp, "")

#### ========================================================================
def PreScan_LWO(fileName=None, **kwd):
	u'''レイヤ簡易走査'''
	(fp, errMsg) = __Check_FileState(fileName)
	if(errMsg):
		print(errMsg),
		return(None)

	(dbg, verb, flag) = DetermineVisOptions(**kwd)

	print(u"走査開始: {}".format(fileName))
	layers = ParseLWO(d=dbg, v=verb, f=flag).ScanLayer(fp)	# 状態表示有り
	fp.close()
	print(u"-------: {} 個のレイヤ".format(len(layers)))

	if(verb):
		print(u"走査完了."),
	return(layers)

#### ========================================================================
def Open_LWO(fileName=None, **kwd):
	u'''ファイルロード'''
	(fp, errMsg) = __Check_FileState(fileName)
	if(errMsg):
		print(errMsg),
		return((None,)*4)

	print(u"読み込み開始: {}".format(fileName))

	(dbg, verb, flags) = DetermineVisOptions(**kwd)
	(tags, layers, surfs, clips) = ParseLWO(d=dbg, v=verb, f=flags).Read_LWO(fp)
	fp.close()

	if(verb):
		print(u"読み込み終了."),
	return(tags, layers, surfs, clips)

#### ========================================================================
class ParseLWO(object):
	__slots__ = (
		"__DBG", "__VERB", "VisFlags",
		"__FnUnpckTAG",
		"__FnUnpckTAGU2",
		"__FnUnpckLAYR",
		"__FnUnpckLWID",
		"__FnUnpckV4",
		"__FnUnpckV3",
		"__FnUnpckCOL",
		"__FnUnpckANG",
		"__FnUnpckF4",
		"__FnUnpckU4",
		"__FnUnpckU2",
		"__FnUnpckU2U2",
		"__FnUnpckUV",
		"__FnUnpckOPAC",
		"__FnUnpckISEQ",
		"Read_VX",
		"Read_nameString",
	)

	#### -----------------------------------------------------------------------
	def __init__(self, **kwd):
		(self.__DBG, self.__VERB, flags) = DetermineVisOptions(**kwd)
		self.VisFlags = VisFlags(flags)

		self.__FnUnpckTAG  = struct.Struct(ReadT.TAG).unpack
		self.__FnUnpckTAGU2= struct.Struct(ReadT.TAGU2).unpack
		self.__FnUnpckLAYR = struct.Struct(ReadT.LAYR).unpack
		self.__FnUnpckLWID = struct.Struct(ReadT.LWID).unpack
		self.__FnUnpckV4   = struct.Struct(ReadT.VEC4).unpack
		self.__FnUnpckV3   = struct.Struct(ReadT.VEC3).unpack
		self.__FnUnpckCOL  = struct.Struct(ReadT.COL3).unpack
		self.__FnUnpckANG  = struct.Struct(ReadT.ANG4).unpack
		self.__FnUnpckF4   = struct.Struct(ReadT.F4).unpack
		self.__FnUnpckU4   = struct.Struct(ReadT.U4).unpack
		self.__FnUnpckU2   = struct.Struct(ReadT.U2).unpack
		self.__FnUnpckU2U2 = struct.Struct(ReadT.U2U2).unpack
		self.__FnUnpckUV   = struct.Struct(ReadT.UV).unpack
		self.__FnUnpckOPAC = struct.Struct(ReadT.OPAC).unpack
		self.__FnUnpckISEQ = struct.Struct(ReadT.ISEQ).unpack

		# py2/3で関数切り替え.
		if(sys.version_info[0] == 2):
			self.Read_VX = self.Read_VX_Py2
			self.Read_nameString = self.Read_nameString_Py2
		else:
			self.Read_VX = self.Read_VX_Py3
			self.Read_nameString = self.Read_nameString_Py3

	#### -----------------------------------------------------------------------
	def setDbg(self, v, *arg):
		self.__DBG = False if(v == None) or (not isinstance(v, int)) else v
		return(self.__DBG)

	#### -----------------------------------------------------------------------
	def getDbg(self, *arg):
		return(self.__DBG)

	#### -----------------------------------------------------------------------
	def setVerbose(self, v, *arg):
		self.__VERB = False if(v == None) or (not isinstance(v, int)) else v
		return(self.__VERB)

	#### -----------------------------------------------------------------------
	def getVerbose(self, *arg):
		return(self.__VERB)

	#### -----------------------------------------------------------------------
	def setFlags(self, flags=0, *arg):
		f = 0 if(flags == None) or (not isinstance(flags, int)) else flags
		self.VisFlags.setFlags(f)
		return(f)

	#### -----------------------------------------------------------------------
	def getFlags(self, *arg):
		return(self.VisFlags.getFlags())

	#### ========================================================================
	#### ========================================================================
	def ScanLayer(self, fp, *arg):
		u'''簡易ファイル走査'''
		layers = []

		while(True):
			try:
				curChunk = chunk.Chunk(fp)	# FORM:size:LWO2の子階層
			except EOFError:
				break

			size = curChunk.chunksize
			if(curChunk.chunkname == ChunkT.LAYR):
				self.Read_Layr(curChunk.read(size), size, layers)
			elif(curChunk.chunkname == ChunkT.VMAP):
				vmap = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG
				if(  vmap == VMapT.MORF):	# Rel Morph
					self.Scan_Morph_VMap(curChunk.read(size), layers[-1].relMorphs)
				elif(vmap == VMapT.SPOT):	# Abs Morph
					self.Scan_Morph_VMap(curChunk.read(size), layers[-1].absMorphs)
				elif(vmap == VMapT.RGB_):
					self.Scan_RGB_VMap(curChunk.read(size), layers[-1])
				elif(vmap == VMapT.RGBA):
					self.Scan_RGBA_VMap(curChunk.read(size), layers[-1])
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.VMAD):
				vmad = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG
				if(  vmad == VMapT.RGB_):
					self.Scan_RGB_VMad(curChunk.read(size), layers[-1])
				elif(vmad == VMapT.RGBA):
					self.Scan_RGBA_VMad(curChunk.read(size), layers[-1])
				elif(vmad == VMapT.WGHT):
					self.Scan_Weight_VMad(curChunk.read(size), layers[-1])
				elif(vmad == VMapT.NORM):
					self.Scan_NORM(curChunk.read(size), layers[-1])
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.VMED):
				vmed = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG
				if(vmed == VMedT.SUBV):
					self.Scan_SUBV_VMed(curChunk.read(size), layers[-1])
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.POLS):
				lwid = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG
				if(  lwid == PolyT.CURV):
					self.Scan_Crvs(curChunk.read(size), layers[-1])
					isExclude = True
				elif(lwid == PolyT.BONE):
					self.Scan_Bones(curChunk.read(size), layers[-1])
					isExclude = True
				else:
					curChunk.skip()

			else:
				curChunk.skip()
		return(layers)

	#### --------------------------------
	def Scan_Morph_VMap(self, data, morphs):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		morphs[name] = 1

	#### --------------------------------
	def Scan_RGB_VMap(self, data, lyr):
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgb):
			lyr.rgb[name] = VMapObject()
		lyr.rgb[name].vmap.update({1:1})

	#### --------------------------------
	def Scan_RGB_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgb):
			lyr.rgb[name] = VMapObject()
		lyr.rgb[name].vmad.update({1:1})

	#### --------------------------------
	def Scan_RGBA_VMap(self, data, lyr):
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgba):
			lyr.rgba[name] = VMapObject()
		lyr.rgba[name].vmap.update({1:1})

	#### --------------------------------
	def Scan_RGBA_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgba):
			lyr.rgba[name] = VMapObject()
		lyr.rgba[name].vmad.update({1:1})

	#### --------------------------------
	def Scan_NORM(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		lyr.normals[name] = 1

	#### --------------------------------
	def Scan_SUBV_VMed(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		lyr.hasEdgeM = True

	#### --------------------------------
	def Scan_Weight_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name == DEF_NAME_EDGE): lyr.hasEdge = True

	#### --------------------------------
	def Scan_Bones(self, data, lyr):
		lyr.bones["dummy"] = 1
		lyr.bonNum = 1

	#### --------------------------------
	def Scan_Crvs(self, data, lyr):
		lyr.crvs = ((1,))
		lyr.crvNum = 1

	#### ========================================================================
	#### ========================================================================
	def Read_LWO(self, fp):
		curPolIdxOffs = 0
		isExclude = False
		(tags, layers, surfs, clips) = ([], [], {}, {})

		while(True):
			try:
				curChunk = chunk.Chunk(fp)	# FORM:size:LWO2の子階層
			except EOFError:
				break

			size = curChunk.chunksize
			if(curChunk.chunkname == ChunkT.TAGS):
				tags = self.Read_Tags(curChunk.read(size), size)####, tags)

			elif(curChunk.chunkname == ChunkT.LAYR):
				curPolIdxOffs = 0
				isExclude = False
				self.Read_Layr(curChunk.read(size), size, layers)

			elif(curChunk.chunkname == ChunkT.CLIP):	# [LAYR]:[CLIP][CLIP]:...
				self.Read_Clips(curChunk.read(size), size, clips)

			elif(curChunk.chunkname == ChunkT.PNTS):	# [LAYR]:[PNTS]:...
				self.Read_Pnts(curChunk.read(size), size, layers[-1])

			elif(curChunk.chunkname == ChunkT.VMAP):	# [LAYR]:[VMPA][VMAP][LWID]:[VMPA][VMAP][LWID]...
				vmap = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG

				if(  vmap == VMapT.WGHT):
					self.Read_Weight_VMap(curChunk.read(size), size, layers[-1].weights)
				elif(vmap == VMapT.MORF):	# Rel Morph
					self.Read_Morph_VMap(curChunk.read(size), size, layers[-1].relMorphs)
				elif(vmap == VMapT.SPOT):	# Abs Morph
					self.Read_Morph_VMap(curChunk.read(size), size, layers[-1].absMorphs)
				elif(vmap == VMapT.TXUV):
					self.Read_UV_VMap(curChunk.read(size), size, layers[-1])
				elif(vmap == VMapT.RGB_):
					self.Read_RGB_VMap(curChunk.read(size), size, layers[-1])
				elif(vmap == VMapT.RGBA):
					self.Read_RGBA_VMap(curChunk.read(size), size, layers[-1])
				elif(vmap == VMapT.PICK):	# Point Selection
					self.Read_Pick_VMap(curChunk.read(size), size, layers[-1].picks)
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.VMAD):	# [LAYR]:[POLS][...][VMPA][VMAD][LWID]:[POLS][...][VMPA][VMAD][LWID]:...
				vmad = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG

				# VMADのpolIndexはPOLS内でのIndex.
				if(  vmad == VMapT.TXUV):
					self.Read_UV_VMad(curChunk.read(size), size, layers[-1], curPolIdxOffs)
				elif(vmad == VMapT.RGB_):
					self.Read_RGB_VMad(curChunk.read(size), size, layers[-1], curPolIdxOffs)
				elif(vmad == VMapT.RGBA):
					self.Read_RGBA_VMad(curChunk.read(size), size, layers[-1], curPolIdxOffs)
				elif(vmad == VMapT.WGHT):
					self.Read_Weight_VMad(curChunk.read(size), size, layers[-1], curPolIdxOffs)
				elif(vmad == VMapT.NORM):
					self.Read_Normal_VMad(curChunk.read(size), size, layers[-1], curPolIdxOffs)
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.VMED):	# [LAYR]:[VMED][LWID]:...
				vmed = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG

				if(  vmed == VMedT.SUBV):
					self.Read_SUBV_VMed(curChunk.read(size), size, layers[-1])
#				elif(vmed == VMedT.EPCK):
#					curChunk.skip()
				else:
					curChunk.skip()

#			elif(curChunk.chunkname == ChunkT.VMPA):

			elif(curChunk.chunkname == ChunkT.POLS):	# [LAYR]:[POLS][LWID]:[POLS][LWID]:...
				lwid = curChunk.read(RSize.TAG)
				size = curChunk.chunksize - RSize.TAG

				if(lwid in { PolyT.FACE, PolyT.PTCH, PolyT.SUBD }):
					isExclude = False
					curPolIdxOffs = self.Read_Pols(curChunk.read(size), size, layers[-1], lwid)

				elif(lwid == PolyT.CURV):
					isExclude = True
					self.Read_Crvs(curChunk.read(size), size, layers[-1])

				elif(lwid == PolyT.BONE):
					self.Read_Bones(curChunk.read(size), size, layers[-1])
					isExclude = True

#				elif(lwid == PolyT.HBLD):	# カスタムポリ:HairBlade:
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.PTAG):	# [LAYR]:[PTAG]size[LWID]size:[PTAG]size[LWID]size:...
				(ptag,) = self.__FnUnpckTAG(curChunk.read(RSize.TAG)) # LWIDでなくstr
				size = curChunk.chunksize - RSize.TAG

				if(ptag == PTagT.SURF):
					if(isExclude):
						isExclude = False
						curChunk.skip()
					else:
						self.Read_PTag_Surf(curChunk.read(size), size, layers[-1], curPolIdxOffs)

				elif(ptag == PTagT.BONE):
					self.Read_PTag_Bone(curChunk.read(size), size, layers[-1], tags)
				elif(ptag == PTagT.BNWT):
					self.Read_PTag_BoneWeightMap(curChunk.read(size), size, layers[-1], tags)
				elif(ptag == PTagT.BNUP):
					self.Read_PTag_BoneUpVec(curChunk.read(size), size, layers[-1], tags)

				elif(ptag == PTagT.SKBM):
					self.Read_PTag_BoneBankMode(curChunk.read(size), size, layers[-1], tags)
				elif(ptag == PTagT.SKBR):
					self.Read_PTag_BoneBankRot(curChunk.read(size), size, layers[-1], tags)

				elif(ptag == PTagT.SKID):
					self.Read_PTag_BoneEditorId(curChunk.read(size), size, layers[-1], tags)
				elif(ptag == PTagT.SKPR):
					self.Read_PTag_BoneEditorParentId(curChunk.read(size), size, layers[-1], tags)

				elif(ptag == PTagT.COLR):
					if(isExclude):
						curChunk.skip()
					else:
						self.Read_PTag_Color(curChunk.read(size), size, layers[-1], tags, curPolIdxOffs)
				elif(ptag == PTagT.PART):
					if(isExclude):
						curChunk.skip()
					else:
						self.Read_PTag_Part(curChunk.read(size), size, layers[-1], tags, curPolIdxOffs)

#				elif(ptag == PTagT.SMGP):
#				elif((ptag == PTagT.HBCA) or (ptag == PTagT.HBCP) or \
#					 (ptag == PTagT.HBUV) or (ptag == PTagT.HBRT)):	# HairBlade
				else:
					curChunk.skip()

			elif(curChunk.chunkname == ChunkT.SURF):
				# [LAYR]:[SURF][SURF]:...
				self.Read_Surf(curChunk.read(size), size, surfs)

			elif(curChunk.chunkname == b"FORM"):	# LWO3 surface
				(tag,) = self.__FnUnpckTAG(curChunk.read(RSize.TAG))
				size = curChunk.chunksize - RSize.TAG

				offs = 0
				if(tag == SBlokT.SURF):
					while(offs < size):	# SURF内縛り
						try:
							curChunk = chunk.Chunk(fp)
						except EOFError:
							break

						offs += curChunk.chunksize

						if(curChunk.chunkname == SBlokT.NAME):
							surf = SurfaceObject()
							(surf.name, nameLen) = self.Read_nameString(curChunk.read(curChunk.chunksize))
							surfs[surf.name] = surf

						elif(curChunk.chunkname == b"FORM"):
							(tag,) = self.__FnUnpckTAG(curChunk.read(RSize.TAG))
							curChunk.skip()
							break

						else:
							curChunk.skip()
				else:
					curChunk.skip()

			else:
				curChunk.skip()

		return(tags, layers, surfs, clips)

	#### ========================================================================
	def Read_nameString_Py2(self, data):
		u'''文字列読取'''
		# "" -> 00 00, idx:0, len:1, %:True, len:2
		# A -> nn 00, idx:1, len:2 %:False, len:2
		# LWID -> 4byte, idx:5, len:6 %:False, len:6
		# 文字が偶数時は+NULLで奇数, paddingで偶数化されてる
		# 文字が奇数時は+NULLで偶数, padding無し
		idx = data.find(b'\0')
		len = idx + (2 - (idx & 0x1))
		name = "" if(idx < 1) else data[:idx]
		return(name, len)

	#### --------------------------------
	def Read_nameString_Py3(self, data):
		u'''文字列読取'''
		idx = data.find(b'\0')
		len = idx + (2 - (idx & 0x1))

		if(idx < 1):
			name = ""
		else:
			# LWO2はcp932/utf8, LWO3はutf8. 何かは不明だから例外を拾って対処.
			try:
				name = data[:idx].decode("utf-8")
			except UnicodeDecodeError:
				name = data[:idx].decode("cp932")
		return(name, len)

	#### ========================================================================
	def Read_VX_Py2(self, data):
		u'''可変長U2/4読取, 先立ってord[0~4]するのはNG. U2だった時は[3~4]は存在しない.'''
		if(ord(data[0]) != 255):	# 2xではdata==str, ordで数値化.
			return(((ord(data[0]) << 8)| ord(data[1])), 2)
		return(((ord(data[1]) << 16)|(ord(data[2]) << 8)| ord(data[3])), 4)

	#### --------------------------------
	def Read_VX_Py3(self, data):
		u'''可変長U2/4読取, 先立ってord[0~4]するのはNG. U2だった時は[3~4]は存在しない.'''
		if(data[0] != 255):			# 3xではbytes, 直接扱う.
			return( ((data[0] << 8)| data[1]), 2)
		return( ((data[1] << 16)|(data[2] << 8)| data[3]), 4)

	#### ========================================================================
	def Read_Tags(self, data, dataSize):
		u'''各種Tag名. [[tag],[tag],...]
		文末の偶数化0埋めの分が空文字として含まれる."ab"\0[pad]"cde"\0"fg"...'''
		# 2xはdata==str, 3xは==bytes,  2xのfilterはlist, 3xはiter.
		# 後工程を考えbytes→str変換.1回/fileなので分岐コストは考えない.
		if(sys.version_info[0] > 2):
			data = data.decode()
		tags = tuple(filter(lambda s: s!="", data.split('\0')))

		self.__DebugPrint_Tags(tags)
		return(tags)

	#### ========================================================================
	def Read_Layr(self, data, dataSize, layers):
		u'''レイヤ情報
		LAYR { number[U2], flags[U2], pivot[VEC12], name[S0], parent[U2] ? }'''
		lyr = LayerObject()

		(lyr.idx, flags) = self.__FnUnpckLAYR(data[:RSize.U4])

		lyr.isHide = (flags & 0x1)	# bitONで不可視
		lyr.piv = self.__FnUnpckV3(data[RSize.U4:RSize.LYRN])		# (x,y,z)
		(name, nameLen) = self.Read_nameString(data[RSize.LYRN:])

		# レイヤ名が未設定の場合, 規定名+(idx+1).を適用.
		lyr.name = name[:] if(name != "") else "{} {}".format(DEF_NAME_LAYER, (lyr.idx +1))

		offs = (RSize.LYRN + nameLen)
		if(dataSize > offs):
			(lyr.parent,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])

		layers.append(lyr)

		self.__DebugPrint_Layer(lyr, flags)

	#### ========================================================================
	def Read_Pnts(self, data, dataSize, lyr):
		u'''頂点配列. [(x,y,z),(...),...]
		PNTS { point-location[VEC12] * }'''
		APD = lyr.pnts.append
		offs = 0
		while(offs < dataSize):
			APD( self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)]) )	# (x,y,z)
			offs += RSize.VEC3

		lyr.pntNum = len(lyr.pnts)

		self.__DebugPrint_Pnts(lyr.name, lyr.pnts, lyr.pntNum)

	#### ========================================================================
	def Read_Pick_VMap(self, data, dataSize, picks):
		u'''Pick(SelectionSet)/VMap
		VMAP { type[ID4], dimension[U2], name[S0], vert[VX]* }
		PICK, dim[U2], name[S0], vert[VX],,,,'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		pick = []
		APD = pick.append
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			APD(pntIdx)

		picks[name] = tuple(pick)	# { name:(pntIdx,,,,) }

		self.__DebugPrint_Pick(name, pick)

	#### ========================================================================
	def Read_Weight_VMap(self, data, dataSize, weights):
		u'''Weight/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		WGHT, dim[U2], name[S0], vert[VX], value[F4]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		wgt = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(wgt[pntIdx],) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			offs += RSize.F4

		weights[name] = wgt

		self.__DebugPrint_Weight_VMap(name, wgt)

	#### ========================================================================
	def Read_Morph_VMap(self, data, dataSize, morphs):
		u'''Morph/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		MORF, dim[U2], name[S0], vert[VX], value[VEC12]'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		mrfs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			mrfs[pntIdx] = self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)])	# (x,y,z)
			offs += RSize.VEC3

		morphs[name] = mrfs

		self.__DebugPrint_Morph(name, mrfs)

	#### ========================================================================
	def Read_RGB_VMap(self, data, dataSize, lyr):
		u'''VertexColor, RGB/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		RGB, dim[U2], name[S0], vert[VX], value[VEC12]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			cols[pntIdx] = self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)]) + (1.0,)	# (r,g,b,A)RGBAにする
			offs += RSize.VEC3

		if(name not in lyr.rgb):
			lyr.rgb[name] = VMapObject()
		lyr.rgb[name].vmap.update(cols)

		self.__DebugPrint_RGBA("RGB/VMap", name, lyr.rgb[name].vmap)

	#### ========================================================================
	def Read_RGBA_VMap(self, data, dataSize, lyr):
		u'''VertexColor, RGBA/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		RGBA, dim[U2], name[S0], vert[VX], value[VEC16]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			cols[pntIdx] = self.__FnUnpckV4(data[offs:(offs + RSize.VEC4)])	# (r,g,b,a)
			offs += RSize.VEC4

		if(name not in lyr.rgba):
			lyr.rgba[name] = VMapObject()
		lyr.rgba[name].vmap.update(cols)

		self.__DebugPrint_RGBA("RGBA/VMap", name, lyr.rgba[name].vmap)

	#### ========================================================================
	def Read_RGB_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''VertexColor, RGB/VMad
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		RGB, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC12])'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			col = self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)]) + (1.0,)	# (r,g,b,A)RGBAにする
			offs += RSize.VEC3
			if(polIdx in cols):
				cols[polIdx][pntIdx] = col
			else:
				cols[polIdx] = dict({ pntIdx: col })

		if(name not in lyr.rgb):
			lyr.rgb[name] = VMapObject()
		lyr.rgb[name].vmad.update(cols)

		self.__DebugPrint_RGBA("RGB/VMad", name, lyr.rgb[name].vmad)

	#### ========================================================================
	def Read_RGBA_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''VertexColor(RGB/RGBA)/VMad
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		RGBA, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC16])'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			col = self.__FnUnpckV4(data[offs:(offs + RSize.VEC4)])	# (r,g,b,a)
			offs += RSize.VEC4
			if(polIdx in cols):
				cols[polIdx][pntIdx] = col
			else:
				cols[polIdx] = dict({ pntIdx: col })

		if(name not in lyr.rgba):
			lyr.rgba[name] = VMapObject()
		lyr.rgba[name].vmad.update(cols)

		self.__DebugPrint_RGBA("RGBA/VMad", name, lyr.rgba[name].vmad)

	#### ========================================================================
	def Read_UV_VMap(self, data, dataSize, lyr):
		u'''UV/VMap. {name:{vmap:{pntIdx:[u,v]}}}
		VMAP { type[ID4], dimension[U2], name[S0], ( vert[VX], value[F4] # dimension )* }
		TXUV, dim[U2], name[S0], vert[VX], value[VEC8]'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		uvs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			uvs[pntIdx] = self.__FnUnpckUV(data[offs:(offs + RSize.UV)])	# (u,v)
			offs += RSize.UV

		if(name not in lyr.uvs):
			lyr.uvs[name] = VMapObject()
		lyr.uvs[name].vmap.update(uvs)

		self.__DebugPrint_UV("VMap", name, lyr.uvs[name].vmap)

	#### ========================================================================
	def Read_UV_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''UV/VMad. {name:{vmad:{polIdx:{pntIdx:[u,v]}}}}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		TXUV,  dim[U2], name[S0],( vert[VX], poly[VX], value[VEC8])'''
		name, nameLen = self.Read_nameString(data[RSize.U2:])

		uvs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			uvVal = self.__FnUnpckUV(data[offs:(offs + RSize.UV)])	# (u,v)
			offs += RSize.UV

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			if(polIdx in uvs):	uvs[polIdx][pntIdx] = uvVal
			else:				uvs[polIdx] = dict( { pntIdx: uvVal} )

		if(name not in lyr.uvs):
			lyr.uvs[name] = VMapObject()
		lyr.uvs[name].vmad.update(uvs)

		self.__DebugPrint_UV("VMad", name, lyr.uvs[name].vmad)

	#### ========================================================================
	def Read_Normal_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''NORM/VMad. {name:{vmad:{polIdx:{pntIdx:[u,v]}}}}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		NORM, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC12])'''
		name, nameLen = self.Read_nameString(data[RSize.U2:])

		norm = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			vec = self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)])	# (x,y,z)
			offs += RSize.VEC3

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex

			if(polIdx in norm):	norm[polIdx][pntIdx] = vec
			else:				norm[polIdx] = dict( { pntIdx: vec} )

		lyr.normals[name] = norm

		self.__DebugPrint_NORM(name, norm)

	#### ========================================================================
	def Read_Weight_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''Weight/VMad. {{id1 id2:val},{id1 id2:val},...}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		WGHT,  dim[U2], name[S0],( vert[VX], poly[VX], value[F4])
		Some info: LW stores a face's points in a clock-wize order (with the normal pointing at you).
		This gives edges a 'direction' which is used when it comes to storing CC edge weight values.
		The weight is given to the point preceding the edge that the weight belongs to.'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name != DEF_NAME_EDGE): return

		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(wgt,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			offs += RSize.F4

			polPnts = lyr.pols[(polIdx + curPolIdxOffs)]	# polIdxはPOLS内でのIndex
			try:
				idx1 = polPnts.index(pntIdx)
			except:
				continue

			pnt2 = polPnts[0] if(idx1 == len(polPnts) -1) else polPnts[idx1 + 1]
			lyr.edgeWgts[(pnt2, pntIdx)] = wgt

		lyr.hasEdge = True

		self.__DebugPrint_EdgeWeight("VMad", name, lyr.edgeWgts)

	#### ========================================================================
	def Read_SUBV_VMed(self, data, dataSize, lyr):
		u'''Modo EdgeWeight. {{id1 id2:val},{id1 id2:val},...}
		VMED { type[ID4], dimension[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )* }
		SUBV, dim[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )*'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(idx1, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(idx2, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(wgt,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			offs += RSize.F4
			lyr.edgeWgtsM[(idx1, idx2)] = wgt

		lyr.hasEdgeM = True

		self.__DebugPrint_EdgeWeight("VMed(Modo)", name, lyr.edgeWgtsM)

	#### ========================================================================
	def Read_Pols(self, data, dataSize, lyr, lwid):
		u'''ポリ配列. [(pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* }'''
		curPolIdxOffs = len(lyr.pols)	# polIdxのoffsetとして.

		pols = []
		PolAPD = pols.append
		offs = 0
		while(offs < dataSize):
			(pntsNum,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			pntsNum &= 0x03FF # 上位6bitはフラグ, max1023
			offs += RSize.U2
			pntsIdxs= []
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [pntIdx,pntIdx,...]
			PolAPD(tuple(pntsIdxs))	# [(pntIdx,pntIdx,...),(...),...]

		lyr.pols.extend(pols)

		self.__DebugPrint_Pols(lyr.name, lwid, pols)

		lyr.polNum += len(pols)	# 累積する

		return(curPolIdxOffs)

	#### ========================================================================
	def Read_Crvs(self, data, dataSize, lyr):
		u'''カーブ配列. [(flag, pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* }'''
		crvs = []
		CrvAPD = crvs.append
		offs = 0
		while(offs < dataSize):
			(val,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			# カーブフラグ NNNN XXNN 0000 0000
			pntsNum = (val & 0x03FF)|((val & 0xF000) >> 2)
			pntsIdxs= [((val & 0x0C00) >> 10), ]
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [flag, pntIdx,pntIdx,...]
			CrvAPD(tuple(pntsIdxs))	# [(flag, pntIdx,pntIdx,...),(...),...]

		lyr.crvs   = tuple(crvs)
		lyr.crvNum = len(crvs)

		self.__DebugPrint_Crvs(lyr.name, crvs, lyr.crvNum)

	#### ========================================================================
	def Read_Bones(self, data, dataSize, lyr):
		u'''ボーン読込
		POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* }'''
		bones = {}
		offs = boneIdx = 0
		while(offs < dataSize):
			offs += RSize.U2 # 頂点数は2固定
			(pntIdx0, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(pntIdx1, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			bones[boneIdx] = BoneObject(pnt0= pntIdx0, pnt1= pntIdx1)
			boneIdx += 1

		lyr.bones  = bones
		lyr.bonNum = len(bones)

		bs = list(range(lyr.bonNum))
		for i in bs:
			bon = bones[i]
			for j in bs:
				if(bon.pnt1 == bones[j].pnt0):	# 端点と始点の結合を親子組と見做す
					bon.childs.append(j)
		# Jointと見做しての親子組の都合上,端点を有意な点と見做す.
		# 親を持たないボーンの始点[0]はダミーとし生成しない.

		self.__DebugPrint_Bone(lyr.name, lyr.bones, lyr.bonNum)

	#### ========================================================================
	def Read_PTag_Bone(self, data, dataSize, lyr, tags):
		u'''ボーン名読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].name = tags[tagIdx]

		self.__DebugPrint_PTagBone("name", lyr.name, lyr.bones, "Bone BONE")

	#### ========================================================================
	def Read_PTag_BoneWeightMap(self, data, dataSize, lyr, tags):
		u'''ボーンウェイト名読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].weight = tags[tagIdx]

		self.__DebugPrint_PTagBone("weight", lyr.name, lyr.bones, "Bone WGHT")

	#### ========================================================================
	def Read_PTag_BoneUpVec(self, data, dataSize, lyr, tags):
		u'''スケルゴン UpVector値読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2

			vstr = tags[tagIdx].split(" ")
			bones[boneIdx].bnup = ( float(vstr[0]),float(vstr[1]),float(vstr[2]) )

		self.__DebugPrint_PTagBone("bnup", lyr.name, lyr.bones, "Bone BNUP")

	#### ========================================================================
	def Read_PTag_BoneEditorParentId(self, data, dataSize, lyr, tags):
		u'''スケルゴン EditorId読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].skpr = int(tags[tagIdx])

		self.__DebugPrint_PTagBone("skpr", lyr.name, lyr.bones, "Bone SKPR")

	#### ========================================================================
	def Read_PTag_BoneEditorId(self, data, dataSize, lyr, tags):
		u'''スケルゴン EditorId読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].skid = int(tags[tagIdx])

		self.__DebugPrint_PTagBone("skid", lyr.name, lyr.bones, "Bone SKID")

	#### ========================================================================
	def Read_PTag_BoneBankMode(self, data, dataSize, lyr, tags):
		u'''スケルゴン BankMode読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].skbm = int(tags[tagIdx])

		self.__DebugPrint_PTagBone("skbm", lyr.name, lyr.bones, "Bone SKBM")

	#### ========================================================================
	def Read_PTag_BoneBankRot(self, data, dataSize, lyr, tags):
		u'''スケルゴン BankRotate値読込'''
		bones = lyr.bones
		offs = 0
		while(offs < dataSize):
			(boneIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2
			bones[boneIdx].skbr = float(tags[tagIdx])

		self.__DebugPrint_PTagBone("skbr", lyr.name, lyr.bones, "Bone SKBR")

	#### ========================================================================
	def Read_PTag_Surf(self, data, dataSize, lyr, curPolIdxOffs):
		u'''Poly&SurfaceIdx. {TagsIdx:[polIdx,polIdx,...]}
		PTAG { type[ID4], ( poly[VX], tag[U2] )* } 
		POL->TAGS(name)->Surface'''
		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2

			if(tagIdx not in lyr.surfTags):
				lyr.surfTags[tagIdx] = []	# 既設でなければ作成

			lyr.surfTags[tagIdx].append(polIdx + curPolIdxOffs)	# POLS内のIndex

		self.__DebugPrint_PTag("SURF", lyr.name, lyr.surfTags)

	#### ========================================================================
	def Read_Surf(self, data, dataSize, surfs):
		u'''Surface
		SURF { name[S0], source[S0], attributes[SUB-CHUNK] * }'''
		surf = SurfaceObject()
		(name, nameLen) = self.Read_nameString(data)
		if(len(name) != 0): surf.name = name

		(srcName, srcNameLen) = self.Read_nameString(data[nameLen:])
		if(len(srcName) != 0): surf.srcName = srcName

		if(self.__DBG and self.VisFlags.SURF): (cnt, tcnt) = (0, 0)

		offs = nameLen + srcNameLen
		while(offs < dataSize):
			(subName, subLen) = self.__FnUnpckTAGU2(data[offs:(offs + RSize.TAGU2)])
			offs += RSize.TAGU2

			if(  subName == SurfT.COLR):		# COLR { base-color[COL12], envelope[VX] }
				(surf.colr) = self.__FnUnpckCOL(data[offs:(offs + RSize.COL3)])
			elif(subName == SurfT.DIFF):		# DIFF { intensity[F4], envelope[VX] }
				(surf.diff,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.LUMI):		# LUMI { intensity[F4], envelope[VX] }
				(surf.lumi,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.SPEC):		# SPEC { intensity[F4], envelope[VX] }
				(surf.spec,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.REFL):		# REFL { intensity[F4], envelope[VX] }
				(surf.refl,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.TRAN):		# TRAN { intensity[F4], envelope[VX] }
				(surf.tran,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.TRNL):		# TRNL { intensity[F4], envelope[VX] }
				(surf.trnl,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.GLOS):		# GLOS { glossiness[F4], envelope[VX] }
				(surf.glos,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
#			elif(subName == SurfT.SHRP):		# SHRP { sharpness[F4], envelope[VX] }
#				(surf.shrp,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.BUMP):		# BUMP { strength[F4], envelope[VX] } 
				(surf.bump,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.SIDE):		# SIDE { sidedness[U2] }
				(surf.side,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			elif(subName == SurfT.SMAN):		# SMAN { max-smoothing-angle[ANG4] }
				(surf.sman,) = self.__FnUnpckANG(data[offs:(offs + RSize.ANG4)])

			elif(subName == SurfT.VCOL):		# VCOL { intensity[FP4], envelope[VX], vmap-type[ID4], name[S0] }
				(intensity,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
				tmp = offs + RSize.F4
				(env, vx) = self.Read_VX(data[tmp:(tmp + RSize.VX)])	# env
				tmp += vx
				(lwid,) = self.__FnUnpckLWID(data[tmp:(tmp + RSize.LWID)])
				tmp += RSize.TAG
				(name, tmp1) = self.Read_nameString(data[tmp:])

				surf.vcol = (name, lwid, intensity)
				# offset本体への積算は不要

			elif(subName == SurfT.NORM):	# NORM { intensity[FP4], name[S0] }
				(surf.norm, nameLen) = self.Read_nameString(data[offs:])

			elif(subName == SurfT.BLOK):	# BLOK { header[SUB-CHUNK], attributes[SUB-CHUNK] * } 
				blok = ImapBlokObject()

				tmp = offs
				tmpMax = offs + subLen
				while(tmp < tmpMax):
					(head, headLen) = self.__FnUnpckTAGU2(data[tmp:(tmp + RSize.TAGU2)])
					tmp += RSize.TAGU2

					if(head == BlokT.IMAP):
						(blok.ord,) = self.__FnUnpckU2(data[tmp:(tmp + RSize.U2)])

						tmp0 = tmp + RSize.U2
						tmp0Max = tmp + headLen
						while(tmp0 < tmp0Max):
							(blk, blkLen) = self.__FnUnpckTAGU2(data[tmp0:(tmp0 + RSize.TAGU2)])
							tmp0 += RSize.TAGU2

							if(  blk == ImgT.CHAN):
								(blok.chan,) = self.__FnUnpckTAG(data[tmp0:(tmp0 + RSize.TAG)])
							elif(blk == ImgT.OPAC):
								(blok.opacType, blok.opac) = self.__FnUnpckOPAC(data[tmp0:(tmp0 + RSize.OPAC)])
							elif(blk == ImgT.ENAB):
								(blok.enable,) = self.__FnUnpckU2(data[tmp0:(tmp0 + RSize.U2)])
							elif(blk == ImgT.NEGA):
								(blok.negative,) = self.__FnUnpckU2(data[tmp0:(tmp0 + RSize.U2)])
							elif(blk == ImgT.AXIS):
								(blok.dispAxis,) = self.__FnUnpckU2(data[tmp0:(tmp0 + RSize.U2)])
							else:
								pass
							tmp0 += blkLen

					elif(head == BlokT.TMAP):
						tmp0 = tmp
						tmp0Max = tmp + headLen
						while(tmp0 < tmp0Max):
							(blk, blkLen) = self.__FnUnpckTAGU2(data[tmp0:(tmp0 + RSize.TAGU2)])
							tmp0 += RSize.TAGU2

							if(  blk == TxtrT.CNTR):
								blok.center = self.__FnUnpckV3(data[tmp0:(tmp0 + RSize.VEC3)])
							elif(blk == TxtrT.SIZE):
								blok.size   = self.__FnUnpckV3(data[tmp0:(tmp0 + RSize.VEC3)])
							elif(blk == TxtrT.ROTA):
								blok.rotate = self.__FnUnpckV3(data[tmp0:(tmp0 + RSize.VEC3)])
#							elif(blk == TxtrT.FALL):
#							elif(blk == TxtrT.OREF):
#							elif(blk == TxtrT.CSYS):
							else:
								pass
							tmp0 += blkLen

					elif(head == BlokT.PROJ):
						(blok.project,) = self.__FnUnpckU2(data[tmp:(tmp + RSize.U2)])
					elif(head == BlokT.AXIS):
						(blok.prjAxis,) = self.__FnUnpckU2(data[tmp:(tmp + RSize.U2)])
					elif(head == BlokT.IMAG):
						(blok.imgIdx, vx) = self.Read_VX(data[tmp:(tmp + RSize.VX)])
					elif(head == BlokT.WRAP):
						(blok.wrapOptW, blok.wrapOptH) = self.__FnUnpckU2U2(data[tmp:(tmp + RSize.U2U2)])
					elif(head == BlokT.WRPW):
						(blok.wrapAmountW,) = self.__FnUnpckF4(data[tmp:(tmp + RSize.F4)])
					elif(head == BlokT.WRPH):
						(blok.wrapAmountH,) = self.__FnUnpckF4(data[tmp:(tmp + RSize.F4)])
					elif(head == BlokT.VMAP):
						(blok.vmapName, nameLen)= self.Read_nameString(data[tmp:])
#					elif(head == BlokT.AAST):
#					elif(head == BlokT.PIXB):
#					elif(head == BlokT.STCK):
#					elif(head == BlokT.TAMP):
					else:
						pass
					tmp += headLen
				surf.imap.append(blok)
				# offset本体への積算は不要

			"""
			elif(subName == SurfT.RFOP):		# RFOP { reflection-options[U2] }
			elif(subName == SurfT.RIMG):		# RIMG { image[VX] }
			elif(subName == SurfT.RSAN):		# RSAN { seam-angle[ANG4], envelope[VX] }
			elif(subName == SurfT.RBLR):		# RBLR { blur-percentage[FP4], envelope[VX] }
				(surf.rblr,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.RIND):		# RIND { refractive-index[F4], envelope[VX] }
				(surf.rind,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.TROP):		# TROP { transparency-options[U2] }
			elif(subName == SurfT.TIMG):		# TIMG { image[VX] }
			elif(subName == SurfT.TBLR):		# TBLR { blur-percentage[FP4], envelope[VX] }
				(surf.tblr,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])
			elif(subName == SurfT.CLRH):		# CLRH { color-highlights[FP4], envelope[VX] }
			elif(subName == SurfT.CLRF):		# CLRF { color-filter[FP4], envelope[VX] }
			elif(subName == SurfT.ADTR):		# ADTR { additive[FP4], envelope[VX] }
			elif(subName == SurfT.GLOW):		# GLOW { type[U2], intensity[F4], intensity-envelope[VX], size[F4], size-envelope[VX] }
			elif(subName == SurfT.LINE):		# LINE { flags[U2], ( size[F4], size-envelope[VX], ( color[COL12], color-envelope[VX] )? )? }
			elif(subName == SurfT.ALPH):		# ALPH { mode[U2], value[FP4] }
			"""

			offs += subLen

		surfs[surf.name] = surf

		self.__DebugPrint_Surf(surf)

	#### ========================================================================
	def Read_PTag_Color(self, data, dataSize, lyr, tags, curPolIdxOffs):
		u'''PTAG_COLR. {"col":[polIdx,,,]}'''
		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2

			tag = tags[tagIdx]	# 色名
			if(tag not in lyr.tagColors):
				lyr.tagColors[tag] = []

			lyr.tagColors[tag].append(polIdx + curPolIdxOffs)

		self.__DebugPrint_PTag("COLR", lyr.name, lyr.tagColors)

	#### ========================================================================
	def Read_PTag_Part(self, data, dataSize, lyr, tags, curPolIdxOffs):
		u'''PTAG_PART. {name:[polIdx,,,]}'''
		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(tagIdx,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])
			offs += RSize.U2

			tag = tags[tagIdx]	# パート名
			if(tag not in lyr.tagParts):
				lyr.tagParts[tag] = []

			lyr.tagParts[tag].append(polIdx + curPolIdxOffs)

		self.__DebugPrint_PTag("PART", lyr.name, lyr.tagParts)

	#### ========================================================================
	def Read_Clips(self, data, dataSize, clips):
		u'''Images, CLIP { idx[U4], attributes[SUB-CHUNK] * }'''
		clip = ClipObject()

		(clip.id,) = self.__FnUnpckU4(data[:RSize.U4])

		offs = RSize.U4
		while(offs < dataSize):
			(subName, subLen) = self.__FnUnpckTAGU2(data[offs:(offs + RSize.TAGU2)])
			offs += RSize.TAGU2

			if(  subName == ClipT.STIL):		# STIL { name[FNAM0] }
				clip.type = subName
				(clip.stillName, tmp) = self.Read_nameString(data[offs:])

			elif(subName == ClipT.ISEQ):		# ISEQ { num-digits[U1], flags[U1], offset[I2], reserved[U2], start[I2], end[I2], prefix[FNAM0], suffix[S0] }
				clip.type = subName
				(clip.seqDigit, clip.seqFlags, clip.seqOffset, clip.seqReserve, clip.seqStart, clip.seqEnd,) =\
					self.__FnUnpckISEQ(data[offs:(offs + RSize.ISEQ)])
				tmp = offs + 10
				(clip.seqName,len0)= self.Read_nameString(data[tmp:])
				tmp += len0
				(clip.seqExt, len0)= self.Read_nameString(data[tmp:])
				# offset本体への積算はNG

			elif(subName == ClipT.ANIM):		# ANIM { filename[FNAM0], server-name[S0], flags[U2], data[...] }
				clip.type = subName
				tmp = offs
				(clip.animName,  len0)= self.Read_nameString(data[tmp:])
				tmp += len0
				(clip.animServer,len0)= self.Read_nameString(data[tmp:])
				tmp += len0
				(clip.animFlags,) = self.__FnUnpckU2(data[tmp:(tmp + RSize.U2)])
				# offset本体への積算はNG

			elif(subName == ClipT.XREF):		# XREF { index[U4], string[S0] }
				clip.type = subName
				(clip.xrefIdx,) = self.__FnUnpckU4(data[offs:(offs + RSize.U4)])
				(clip.xrefName, tmp)= self.Read_nameString(data[(offs + RSize.U4):])

#			elif(subName == ClipT.STCC):		# STCC { lo[I2], hi[I2], name[FNAM0] }
#				clip.type = subName
#				(surf.refl,) = self.__FnUnpckF4(data[offs:(offs + RSize.F4)])

			elif(subName == ClipT.TIME):		# TIME { start-time[FP4], duration[FP4], frame-rate[FP4] }
				(clip.stTime, clip.duration, clip.fps) = self.__FnUnpckV3(data[offs:(offs + RSize.VEC3)])
			elif(subName == ClipT.NEGA):		# NEGA { enable[U2] }
				(clip.neg,) = self.__FnUnpckU2(data[offs:(offs + RSize.U2)])

#			elif(subName == ClipT.CONT):		# CONT { contrast-delta[FP4], envelope[VX] }
#			elif(subName == ClipT.BRIT):		# BRIT { brightness-delta[FP4], envelope[VX] }
#			elif(subName == ClipT.SATR):		# SATR { saturation-delta[FP4], envelope[VX] }
#			elif(subName == ClipT.HUE_):		# HUE { hue-rotation[FP4], envelope[VX] }
#			elif(subName == ClipT.GAMM):		# GAMM { gamma[F4], envelope[VX] }
#			elif(subName == ClipT.IFLT):		# IFLT { server-name[S0], flags[U2], data[...] }
#			elif(subName == ClipT.PFLT):		# PFLT { server-name[S0], flags[U2], data[...] }
#			elif(subName == ClipT.FLAG):		# FLAG { flags[U4] }
			offs += subLen

		clips[clip.id] = clip

		self.__DebugPrint_Clip(clip)

	#### ========================================================================
	def __DebugPrint_Pnts(self, name, pnts, pntNum):
		if(self.__DBG and self.VisFlags.PNT):
			msg = [ u"{:4}{}: {}".format("", i, v) for (i, v) in enumerate(pnts) ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"  PNTS読込 ( {} ) total: {} pts".format(name, pntNum)),
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Pols(self, name, lwid, pols):
		if(self.__DBG and self.VisFlags.POL):
			msg = [ u"{:4}{}: {}".format("", i, v) for (i, v) in enumerate(pols) ]
		else:
			msg = []

		if(self.__VERB):
			if(not isinstance(lwid, str)):	# py3にbasestringは無いのでstrで判定, py3ではbytes.
				lwid = lwid.decode()
			msg.insert(0, u"  {}読込 ( {} ) total: {} pols".format(lwid, name, len(pols)))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Crvs(self, name, crvs, crvNum):
		if(self.__DBG and self.VisFlags.CRV):
			msg = [ u"\t\t{}: {}".format(i, v) for (i, v) in enumerate(crvs) ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"  CURV読込 ( {} ) total: {} crvs".format(name, crvNum))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Pick(self, name, pick):
		if(self.__VERB):
			print(u"{:4}PICK/VMap ( {} )\n".format("", name))
		if(self.__DBG and self.VisFlags.SELS):
			print("{:6}{}\n".format("", pick))

	#### --------
	def __DebugPrint_Weight_VMap(self, name, wgt):
		if(self.__VERB):
			print(u"{:4}WGHT/VMap ( {} )\n".format("", name))
		if(self.__DBG and self.VisFlags.WGHT):
			print(u"{:6}{}\n".format("", wgt))

	#### --------
	def __DebugPrint_EdgeWeight(self, name, tag, wgt):
		if(self.__DBG and self.VisFlags.EDGE):
			msg = [ u"{:6}{}".format("", wgt) ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}EdgWgt/{} ( {} )".format("", tag, name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Morph(self, name, vmapDict):
		if(self.__DBG and self.VisFlags.MORF):
			msg = [ u"{:6}{}: {}".format("", k, v) for (k, v) in sorted(vmapDict.items()) ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}MORF/VMap ( {} )".format("", name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_RGBA(self, tag, name, vmapDict):
		if(self.__DBG and self.VisFlags.VCOL):
			msg = [ u"{:6}{}: {}".format("", k, v) for (k, v) in vmapDict.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}{} ( {} )".format("", tag, name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_UV(self, tag, name, vmapDict):
		if(self.__DBG and self.VisFlags.UV):
			msg = [ u"{:6}{}: {}".format("", k, v) for (k, v) in vmapDict.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}UV/{} ( {} )".format("", tag, name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_NORM(self, name, vmapDict):
		if(self.__DBG and self.VisFlags.NORM):
			msg = [ u"{:6}{}: {}".format("", k, v) for (k, v) in vmapDict.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}NORM/Vmad ( {} )".format("", name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Bone(self, name, bones, bonNum):
		if(self.__DBG and self.VisFlags.BONE):
			msg = [ u"{:4}{}: ({}, {}) {}".format("", k, bo.pnt0, bo.pnt1, bo.childs) for (k, bo) in bones.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"  Bone読込 ( {} ) total: {} bones".format(name, bonNum))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_PTagBone(self, tag, name, bones, typeName):
		if(self.__DBG and self.VisFlags.BONE):
			msg = [ u"{:6}{}: {}".format("", k, bo[tag]) for (k, bo) in bones.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}PTag {}( {} )".format("", typeName, name))
			msg.append(u"{:6}End of PTag: {}".format("", typeName))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_PTag(self, tag, name, dat):
		if(self.__DBG and self.VisFlags.PTAG):
			msg = [ u"{:6}{}: {}".format("", k, v) for (k, v) in dat.items() ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"{:4}PTag {}( {} )".format("", tag, name))
			msg.append(u"{:6}End of PTag: {}".format("", tag))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Surf(self, surf):
		msg = []
		if(self.__DBG):
			if(self.VisFlags.SURF and self.VisFlags.IMAP):
				msg = []
				APD = msg.append
				for x in surf.__slots__:
					if(x != "imap"):
						APD(u"  {}: {}".format(x, surf[x]))
					else:
						for (i ,im) in enumerate(surf.imap):
							APD("  imap: [{}]".format(i))
							for z in im.__slots__:
								APD(u"{:4}{}: {}".format("", z, surf.imap[i][z]))
			elif(self.VisFlags.SURF):
				msg = [ u"  {}: {}".format(x, surf[x]) for x in surf.__slots__ if(x != "imap") ]

		if(self.__VERB):
			msg.insert(0, u"SURF読込 ( {} )".format(surf.name))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Clip(self, clip):
		if(self.__DBG and self.VisFlags.CLIP):
			msg = [ u"  {}: {}".format(x, clip[x]) for x in clip.__slots__ ]
		else:
			msg = []

		if(self.__VERB):
			msg.insert(0, u"Clip読込 ( {} )".format(clip.id))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Layer(self, lyr, flags):
		if(self.__VERB):
			msg = [u"レイヤ {} を処理".format(lyr.idx)]
		else:
			msg = []

		if(self.__DBG and self.VisFlags.LAYR):
			msg.append(u"  Name: {}".format(lyr.name))
			msg.append(u"  LayerIdx: {} (Parent: {})".format(lyr.idx, lyr.parent))
			msg.append(u"  flag: {:x}".format(flags))
			msg.append(u"  pivot: {}".format(lyr.piv))
		if(msg): print(u"\n".join(msg+[u"\n"]))

	#### --------
	def __DebugPrint_Tags(self, tags):
		if(self.__DBG and self.VisFlags.TAGS):
			msg = [ u"  {},".format(a) for a in tags ]
			print(u"\n".join([u"TAGS:"] + msg +[u"\n"]))
